from __future__ import annotations

import pandas as pd
import numpy as np
import copy
import scipy.signal as ss
import warnings

# Type annotations without circular import
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from lab3.experiment import BehaviorExperiment

from lab3.core import Item, Group

from .base import EventsFile, Event, OnsetEvent, OffsetEvent, IntervalEvent, bool2events
from .group import BehaviorEventGroup

class SpatialEvent(Event):
    def __init__(self, *args, **kwargs):
        """
        
        TO DOCUMENT

        """
        super().__init__(*args, **kwargs)

        if self.duration > 0.5:
            warnings.warn(f"Creating a SpatialEvent that spans {self.duration*100:.0f}% the belt, "
                    "are you sure this is what you want to do?")

    def to_bin(self, bins):
        bin_on = np.searchsorted(bins, self.onset, side='right') - 1
        if np.isnan(self.offset):
            bin_off = np.nan
        else:
            bin_off = np.searchsorted(bins, self.offset, side='right') - 1
            
        return self.__class__(onset=bin_on, offset=bin_off)
    
    def to_time_events(self, expt, name=None):
        if name is None:
            name = f"position_({self.onset}, {self.offset})"
        
        positions = expt.format_behavior_data()['treadmillPosition']
        bool_array = (self.onset < positions) & (positions < self.offset)
        return BehaviorEventGroup(bool2events(bool_array), name=name)

class SpatialEventGroup(BehaviorEventGroup):
    @classmethod
    def load(cls, expt : BehaviorExperiment, behavior_key : str, **kwargs):
        events = super().load(expt, behavior_key, as_time_events=False, **kwargs)
        positions = pd.Series(expt.format_behavior_data()['treadmillPosition'], name='positions')
        onsets = positions.loc[events.onset]
        offsets = positions.loc[events.offset]
        
        return cls([SpatialEvent(onset=on, offset=off) 
                    for on, off in zip(onsets, offsets)], name=behavior_key)
    
    def positions2bins(self, nbins=None, bins=None):
        assert bins or nbins
        if bins is None:
            bins = np.linspace(0, 1, nbins)
        
        return self.__class__([event.to_bin(bins) for event in self], 
                              name=self.name)
    
    def to_time_events(self, expt):
        nframes = len(expt.imaging_times())
        bool_array = np.sum([spatial_event.to_time_events(expt).bool_array(nframes) 
                             for spatial_event in self], axis=0)
        
        return BehaviorEventGroup(bool2events(bool_array), name=self.name)
        



import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages


class PdfPlotter:

    def __init__(self, filepath, closefigs=True):
        if closefigs:
            for fignum in plt.get_fignums():
                plt.close(fignum)

        self.plot_number = 0
        self.filepath = filepath
        self.figs = []
        if '_pdf_plotter' not in plt.__dict__.keys():
            plt._figure = plt.figure
            plt._show = plt.show
        plt.figure = self.figure
        plt.show = self.save
        plt._pdf_plotter = self

    def save(self):
        pp = PdfPages(self.filepath)
        for figure in self.figs:
            plt._figure(figure.number)
            plt.savefig(pp, format='pdf')
        pp.close()

    def figure(self, num=None, **kwargs):
        if num is not None:
            return plt._figure(num=num, **kwargs)

        fig = plt._figure(**kwargs)
        self.figs.append(fig)
        self.plot_number = 0

        return fig



from __future__ import annotations

import numpy as np
from scipy.ndimage import gaussian_filter1d
import scipy.stats as stats 
import pandas as pd
import itertools as it

from typing import TYPE_CHECKING

from lab3.core import Analysis
from lab3.core.filters import IndexIdentity, ColumnIdentity
from lab3.analysis.base import SignalAnalysis
from lab3.event import Event, EventGroup, LapEvent, LapGroup

from lab3.core.helpers import save_load

if TYPE_CHECKING:
    from lab3.experiment import ImagingExperiment

from . import SimpleSpatialTuning, SimplePFDetector, CircularSimpleSpatialTuning

def discretize_position(position, nbins=100):
    bins = np.linspace(0,1,nbins+1)[:-1]
    return np.searchsorted(bins, position) - 1

def discretize_laps(laps, nmax):
    out = np.ones(nmax)*(len(laps)-1)
    for i in range(1,len(laps)):
        out[laps[i-1]:laps[i]] = i-1

    return out

# TODO: Reconcile with abstract spatial tuning
class PlaceFields(SignalAnalysis):
    #name = 'PlaceFields'
    constructor_ids = ['num_bins', 'num_shuffles', 
                       'pval_threshold', 'min_trials_active', 'sigma']
    def to_columns(self):
        return pd.Index(['field_start', 'field_stop'])
    
    def __init__(self, num_shuffles=100, num_bins=100, pval_threshold=0.01, 
                 min_trials_active=3, sigma=3, **kwargs):
        """A shuffle-based place field identification algorithm. For each cell, 
        we draw surrogate tuning curves from a null distribution of shuffled
        signals. Locations where the true tuning curve exceeds some percentile 
        (default: 99th) of the shuffle distribution are identified as (statistically 
        significant) place fields.

        NB: Since this is a randomized (shuffle-based) algorithm, different runs may return 
        slightly different results.

        Initialization parameters
        ----------
        num_shuffles : int
            Number of shuffles to use to construct null distribution.
            The more the slower. Default: 100.
        num_bins : int
            Number of bins to discretize the track. Default: 100. 
        pval_threshold : float, optional
            Threshold to call a "significant" place field (default p<0.01)
        min_trials_active : int, optional
            Minimum number of laps a "place field" must occur on to be called a place field
        sigma : float, optional
            Standard deviation of the gaussian smoothing kernel applied to the
            tuning curves. By default, mild (3 bin) smoothing is applied.

        Examples
        --------
        >>> # Identify place fields
        >>> strategy = PlaceFields()
        >>> signal_spec = {"signal_type": "dfof", "label": "suite2p"}
        >>> place_fields = strategy.apply_to(
        >>>     expt, signal_spec=signal_spec)
        """

        super().__init__(num_shuffles=num_shuffles, num_bins=num_bins, 
                         pval_threshold=pval_threshold, 
                         min_trials_active=min_trials_active, sigma=sigma, 
                         **kwargs)
        self.num_shuffles = num_shuffles
        self.num_bins = num_bins
        self.pval_threshold=pval_threshold
        self.sigma = sigma
        self.min_trials_active = min_trials_active
        self.tuning_strategy = CircularSimpleSpatialTuning(sigma=sigma)
    
    @save_load
    def apply_to(self, expt : ImagingExperiment, signal_spec, roi_filter=IndexIdentity, 
                    time_filter=ColumnIdentity):
        laps = expt.format_behavior_data()['lap']
        position = expt.format_behavior_data()['treadmillPosition']
        signals = self._load_signals(expt, signal_spec, roi_filter=roi_filter, 
                                        time_filter=time_filter)
        pfs = self.calculate(signals, laps, position)
        pfs.columns = self.columns
        
        return pfs
    
    def calculate(self, signals, laps, position):
        discrete_position = discretize_position(position, nbins=self.num_bins)
        discrete_laps = discretize_laps(laps, len(discrete_position))

        if 0 < self.min_trials_active < 1.:
            min_trials = int(len(laps) * self.min_trials_active)
        else:
            min_trials = self.min_trials_active
        
        try:
            place_fields = SimplePFDetector(self.tuning_strategy, num_shuffles=self.num_shuffles, 
                                            pval_threshold=self.pval_threshold, 
                                            min_trials_active=min_trials).detect(
                                            signals, discrete_position, laps=discrete_laps)
        except:
            place_fields = None
        
        return place_fields




import pandas as pd
import pprint
import os
import json
import pickle as pkl
from functools import reduce
import inspect
import warnings

# O(|MRO_1|+|MRO_2|+|MRO_3|+...)
def infer_item_class(iterable_of_items, typecheck=False):
    if len(iterable_of_items) == 0:
        return None

    classes = [item.__class__.mro() for item in iterable_of_items]

    if typecheck and len(set(map(tuple, classes))) != 1:
        raise TypeError('All items in group must be of same type!')

    return infer_lowest_common_ancestor(classes)


def infer_lowest_common_ancestor(list_of_classes):
    if len(list_of_classes) == 1:
        return list_of_classes[0][0]
    else:
        shared = common_ancestors(list_of_classes[0], list_of_classes[1])
        return infer_lowest_common_ancestor([shared] + list_of_classes[2:])


def common_ancestors(ancestors1, ancestors2):
    ancestor_set = set(ancestors2)
    shared = []
    for a in ancestors1:
        if a in ancestor_set:
            shared.append(a)
    return shared

def wrap(df, key, header=None):
    return pd.concat([df], keys=[key], names=[header])

def bind_args(cls, args, kwargs):
    return cls(*args, **kwargs)

def check_instance(obj, cls):
    try:
        return isinstance(obj, cls)
    except TypeError:
        from lab3.core.registry import lookup_class
        cls = lookup_class(cls)
        return isinstance(obj, cls)

def is_target_object(arg, function, name=None):
    """Checks if `arg` is the target object of `function`
    based on static type hints in the function signature
    Examples
    --------
    >>> def foo(a: int):
    >>>     return a
    >>> is_target_object(foo, 4) # returns True
    >>> is_target_object(foo, 'attila') # returns False

    If the function is not type-annotated, returns `None`
    with a warning:
    >>> def bar(b):
    >>>     return b
    >>> is_target_object(bar, 4) # returns None

    Returns
    --------
    {True, False} or None if untyped
    """
    signature = inspect.signature(function)
    params = signature.parameters
    keys = list(params.keys())

    if keys[0] != 'self':
        first_arg = params[keys[0]]
    else:
        first_arg = params[keys[1]]

    if first_arg.annotation is inspect._empty:
        if name is None:
            name = function.__name__
        warnings.warn(f"The Analysis `{name}` is not type-annotated "
                      "(consider adding annotations to allow type-checking)")
        return None
    elif check_instance(arg, first_arg.annotation):
        return True
    else:
        return False

def conjunction(list_of_filters):
    return lambda df: reduce(lambda x,y: x & y, 
                        [filt(df) for filt in list_of_filters])

def disjunction(list_of_filters):
    return lambda df: reduce(lambda x,y: x | y, 
                        [filt(df) for filt in list_of_filters])

def _params2str(raw_params):
    params = {}
    for key, value in raw_params.items():
        if isinstance(value, dict):
            params[key] = _params2str(value)
        else:
            params[key] = str(value)
    return params

def stringify_params(raw_params):
    return json.dumps(_params2str(raw_params), sort_keys=True)

class ResultStore(pd.HDFStore):
    def __init__(self, path, hash_size=16, **kwargs):
        self.hash_size = hash_size
        self.h5_path = os.path.join(path, 'analysis_results.h5')
        self.pkl_path = os.path.join(path, 'result_parameters.pkl')
        
        try:
            with open(self.pkl_path, 'rb') as f:
                self._pkl_data = pkl.load(f)
        except (EOFError, FileNotFoundError, IOError):
            self._pkl_data = {}        
        
        super().__init__(self.h5_path, mode='a', **kwargs)
        
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        
    def close(self):
        try:
            self._pkl_handle.close()
        except AttributeError:
            pass
            
        super().close()
        
    def get_hash(self, parameters):
        return str(hex(hash(stringify_params(parameters)) % 2**self.hash_size))
    
    def _update_pkl(self, data):
        with open(self.pkl_path, mode='wb+') as f:
            pkl.dump(data, f)
    
    def info(self):
        info = super().info()
        
        legend = stringify_params(self._pkl_data)
        
        return info + "\nLEGEND\n" + pprint.pformat(legend)
        
    def put(self, parameters, dataframe):
        hash_val = self.get_hash(parameters)            
        self._pkl_data[hash_val] = parameters
        self._update_pkl(self._pkl_data)
        
        super().put(hash_val, pd.DataFrame(dataframe))
    
    def get(self, parameters):
        hash_val = self.get_hash(parameters)
        return super().get(hash_val)
    
def save_load(apply_to):
    def save_loadable_apply_to(self, expt, verbose=False, **runtime_parameters):
        try:
            if self.load_saved:
                return self.load(expt, verbose=self.verbose, **runtime_parameters)
        except Exception as e:
            pass
            
        result = apply_to(self, expt, **runtime_parameters)
        
        try:
            if self.save_results and not self.loaded:
                self.save(result, expt, verbose=self.verbose, **runtime_parameters)
        except Exception as e:
            pass        
        return result
    
    return save_loadable_apply_to
    



"""Classes that extend the functionality of experiment objects to handle
operations specific to experiments collected on the virtual reality rigs."""

import os

import numpy as np
from .base import BehaviorExperiment, ImagingExperiment


class VirtualBehaviorMixin:

    @property
    def vr_context_names(self):
        """Get the name of variables in self.behavior_data that correspond
        to virtual reality contexts."""

        scene_names = []
        for ctx_dict in \
                self.behavior_data['__trial_info']['contexts'].values():

            if 'vr_file' in ctx_dict:
                scene_names.append(os.path.splitext(ctx_dict['vr_file'])[0])
            elif 'scene_name' in ctx_dict:
                scene_names.append(ctx_dict['scene_name'])
            else:
                # not a VR context. any other cases we should handle?
                pass

        return sorted(scene_names)

    def vr_context_vars(self, **kwargs):
        return {name: self.format_behavior_data(**kwargs)[name]
                for name in self.vr_context_names}

    def iti(self, **kwargs):
        """Use context variables to determine sample bins that fall during
        the inter-trial-interval periods"""
        iti = np.sum(list(self.vr_context_vars(**kwargs).values()),
                     axis=0) == 0

        # correct the right boundaries of the ITIs
        lap_starts = np.where(np.diff(iti.astype('int')) == -1)[0] + 1
        for lap_num, s in enumerate(lap_starts):
            margin = np.argmin(self.discrete_position(**kwargs)[s:s + 30])
            iti[s:s + margin] = True

            # there is an unhandled edge-case here -- what if the animal runs
            # really fast at the start of the lap and so misses pos 0, but
            # then back-pedals?

        # iti_ends = np.where(np.diff(iti.astype('int')) == -1)[0] + 1
        # for end in iti_ends:
        #     # look ahead 20 samples, find the first sample that drops below 10
        #     pos_snippet = self.discrete_position(**kwargs)[end:end + 20]
        #     extension = np.where(pos_snippet < 10)[0][0]  # + 1
        #     iti[end:end + extension] = True

        return iti

    def sample_vr_context_id(self, **kwargs):
        """Get a list with elements corresponding to the context name on each
        sample."""
        sample_id = np.asarray([None] * len(self.iti(**kwargs)))
        for key, value in self.vr_context_vars(**kwargs).items():
            sample_id[value] = key
        sample_id[self.iti(**kwargs)] = None
        return sample_id

    def lap_vr_context_id(self, **kwargs):
        """Get a list with elements corresponding to the context name on each
        lap"""
        sample_ids = self.sample_vr_context_id(**kwargs)[~self.iti(**kwargs)]

        true_lap_starts = np.concatenate(
            [[0], np.where(np.diff(
                self.laps(**kwargs)[~self.iti(**kwargs)]))[0] + 2]
            )

        return sample_ids[true_lap_starts]

    def valid_samples(self, velocity_threshold=5, **kwargs):
        """Make a boolean variable that is True at each sample that exceeds
        the velocity threshold and is not during an ITI period

        Parameters
        ----------
        velocity_threshold : float, optional
            Minimum velocity for inclusion, in cm/sec. Defaults to 5.
        """
        return (~ self.iti(**kwargs)) \
            & (self.velocity(**kwargs) > velocity_threshold)

    #
    # def format_behavior_data(self, **kwargs):
    #     """Adds ITI and valid_samples variables to the behavior_data
    #     dictionary"""
    #     raise NotImplementedError


class VirtualImagingMixin:

    pass


class VirtualBehaviorExperiment(VirtualBehaviorMixin, BehaviorExperiment):

    pass


class VirtualImagingExperiment(VirtualImagingMixin, VirtualBehaviorMixin,
                               ImagingExperiment):

    pass



import sys


class ProgressBar:
    """ Simple command line progress bar class

    Parameters
    ----------
    max_val : int
        the number of cycles equivalent to 100% progress
    display_length : int, optional
        the length, in number of characters, to display the progress bar
    marger : char, optional
        the character to use to represent progress


    Example
    -------
    >>> from lab.misc.progressbar import ProgressBar
    >>> import time
    >>> num_iterations = 100
    >>> progress = ProgressBar(num_iterations)
    >>> for i in xrange(num_iterations)
    >>>     progress.update(i)
    >>>     time.sleep(1)
    >>> progress.end()
    """

    def __init__(self, max_val, display_length=20, marker='=', annotation=''):
        self._display_length = display_length
        if max_val == 0:
            max_val = 1
        self._max_val = max_val
        self._marker = marker
        self._annotation = annotation

    def update(self, i):
        """Update the progress bar. Overwrites the current line in order to
        appear as moving accross

        Parameters
        ----------
        i : int
            The current cycle number/iteration to represent
        """

        sys.stdout.write('\r')
        per_comp = i * 1.0 / self._max_val
        text_str = "\t{}[%-{}s] %d%%".format(
            self._annotation, self._display_length)
        sys.stdout.write(
            text_str % (self._marker*int(per_comp*self._display_length),
                        per_comp * 100))
        sys.stdout.flush()

    def end(self):
        """ Ends the progress bar, which is set to 100%.
        Prints a new line character"""

        self.update(self._max_val)
        print('')



"""Methods for calculating dF/F. Implement custom strategies by subclassing
DFOFStrategy and filling in the details."""

from abc import abstractmethod
# from .base import BaseSignalTransformer

import numpy as np
import pandas as pd
# import multiprocessing as mp

from lab3.signal import filters
from lab3.core import Automorphism


class DFOFStrategy(Automorphism):

    """Base class for dF/F calculation strategies. Derived strategies should
    implement the _filter_final_signal and _calculate_baseline methods."""

    name = "DFOF"

    def apply_to(self, experiment, channel='Ch2', label=None, **kwargs):
        """Apply dF/F analysis to an experiment object.

        Parameters
        ----------
        experiment : lab3.experiment.base.ImagingExperiment
            Instance of ImagingExperiment to analyze
        channel :
        label :

        Additional keyword arguments are passed directly to `calculate`.
        """
        sig_dict = self._load_signals(experiment, channel=channel, label=label)
        return self.calculate(**sig_dict, **kwargs)

    def calculate(self, signal, *args, **kwargs):
        """Calculate dF/F

        Parameters
        ----------
        signal : array-like (n_rois, n_samples)
            2D array of roi signals

        Returns
        -------
        dfof : pd.DataFrame (n_rois, n_samples)
            2D array of roi dF/F
        """

        if len(signal.shape) != 2:
            raise ValueError("Input signal must be 2D")

        try:
            del self.baseline
        except AttributeError:
            pass

        base = self.calculate_baseline(signal, *args, **kwargs)
        return self._filter_final_signal(
            (signal - base) / base, *args, **kwargs)

    # ----------------- Implement these methods to subclass ----------------- #

    @abstractmethod
    def _filter_final_signal(self, signal):
        """Apply smoothing to the final trace."""

    @abstractmethod
    def _calculate_baseline(self, signal):
        """Estimate baseline for dFOF calculation."""

    def calculate_baseline(self, signal, *args, **kwargs):
        """
        Parameters
        ----------
        signal : array-like (n_rois, n_samples)
            2D array of roi signals

        Returns
        -------
        baseline : pd.DataFrame (n_rois, n_samples)
            2D array of roi baselines
        """
        try:
            return self.baseline
        except AttributeError:
            self.baseline = self._calculate_baseline(signal, *args, **kwargs)
            return self.baseline

    # ------------------------------- Optional ------------------------------ #

    def _load_signals(self, experiment, channel, label):
        """By default only the raw signal dataframe is passed to `calculate`
        when using `apply`. If you override `calculate` to take additional
        arguments in your custom strategy, you should load and add them to the
        dictionary here, which will be unpacked as arguments to `calculate`.
        """
        sig_dict = {
            'signal': experiment.signals(signal_type='raw', channel=channel,
                                         label=label, max_frame=None)
        }
        return sig_dict


class SlowTrendMixin:

    """Mixin class that provides a final detrending step at the end of
    dF/F calculation, calculated as a rolling median filter."""

    # TODO should this override the calculate method? It would make composition
    # easier

    def calculate_slow_trend(self, signal, window=None, min_periods=None):
        """
        Parameters
        ----------
        signal : array-like (n_rois, n_samples)
            2D array of roi signals
        window : int, optional
            Size of window in frames to compute rolling median
        min_periods : float (0, 1], optional
            Fraction of frames in rolling window that must be non-NaN to return
            a non-NaN value in the filtered signal

        Returns
        -------
        slow_trend : pd.DataFrame (n_rois, n_samples)
            2D array of roi slow trends (rolling median)
        """
        try:
            return self.slow_trend
        except AttributeError:
            if window is None:
                return pd.DataFrame(np.zeros(signal.shape))
            else:
                self.slow_trend = pd.DataFrame(signal).rolling(
                    window=window, min_periods=int(window * min_periods),
                    center=True, axis=1).median()
                return self.slow_trend


class NaNFillMixin:

    """Mixin class to add NaN-filling functionality to dF/F calculation."""

    # TODO this should override the calculate method and fill in NaNs
    # How to handle extra parameters?

    pass


class QuantileDFOF(DFOFStrategy):

    """Basic dF/F, with baseline calculated as a rolling percentile of the
    smoothed raw trace. No filtering is applied to the trace in the final
    calculation.

    Parameters
    ----------
    quantile : float (0, 1], optional
        Quantile for baseline calculation. Defaults to 0.08
    window : int, optional
    sigma : float, optional
        Standard deviation (in frames) of gaussian smoothing kernel, applied
        prior to the quantile filter in the baseline calculation.
        Defaults to 5.
    min_periods : float (0, 1], optional

    Attributes
    ----------
    baseline : pd.DataFrame
    """

    def __init__(self, quantile=0.08, window=300, sigma=5, min_periods=0.2):
        self.quantile = quantile
        self.window = window
        self.sigma = sigma
        self.min_periods = min_periods

    def _filter_final_signal(self, signal):
        """No final filtering"""
        return signal

    def _calculate_baseline(self, signal):
        """Smooth signal with a Gaussian, then apply rolling quantile filter"""

        kwargs = {'window': self.window,
                  'min_periods': int(self.window * self.min_periods),
                  'center': True,
                  'axis': 1}

        smooth_df = pd.DataFrame(signal).rolling(
            win_type='gaussian', **kwargs).mean(std=self.sigma)

        return pd.DataFrame(smooth_df).rolling(**kwargs).\
            quantile(self.quantile)


class CustomFilterDFOF(DFOFStrategy):

    """Pass functions for baseline and filter methods and their respective
    keyword arguments.

    Parameters
    ----------
    base_func, filt_func : function
        A function that takes a 2D array of signals and returns a 2D array of
        baselines (base_func), final filtered signals (filt_func)
    base_kws, filt_kws : dict, optional
        Additional keyword arguments for base_func and filt_func, stored as
        key, value pairs in a dictionary. Optional.

    Attributes
    ----------
    baseline : pd.DataFrame
    """

    def __init__(self, base_func, filt_func, base_kws={'axis': 1},
                 filt_kws={'axis': 1}):
        self._filter_final_signal = lambda x: filt_func(x, **filt_kws)
        self._calculate_baseline = lambda x: base_func(x, **base_kws)


class JiaDFOF(DFOFStrategy, SlowTrendMixin):

    """Implements the dF/F calculation method of Jia et al 2010 (Nat. Protocols)
    with additional options to detrend slow changes.

    Parameters
    ----------
    t1 : int, optional
        Window size in frames for initial smoothing. A good value is ~3 sec
        of imaging time. Defaults to 90 frames.
    t2 : int, optional
        Window size in frames for rolling minimum baseline. A good value is ~60
        sec of imaging time. Defaults to 1800 frames.
    exp : float, optional
        Time constamt for exponential filtering of final trace.
        Not implemented
    slow_trend_window : int, optional
        If passed, the final trace is median-filtered in a rolling window of
        this size to remove slow changes (e.g. due to photobleaching). This
        is conventionally larger than the baseline window. Defaults to None
        (no detrending is applied).
    min_periods_t1, min_periods_t2, min_periods_slow : float (0, 1], optional
        Fraction of frames in rolling t1, t2, and slow_trend windows that must
        be non-NaN to return a non-NaN value.

    Attributes
    ----------
    baseline : pd.DataFrame
    slow_trend : pd.DataFrame
    """

    def __init__(self, t1=90, t2=1800, exp=None, min_periods_t1=0.2,
                 min_periods_t2=0.2, slow_trend_window=None,
                 min_periods_slow=0.2):
        self.t1 = t1
        self.t2 = t2
        self.exp = exp
        self.min_periods_t1 = min_periods_t1
        self.min_periods_t2 = min_periods_t2
        self.slow_trend_window = slow_trend_window
        self.min_periods_slow = min_periods_slow

    def _filter_final_signal(self, signal):
        """Filter with exponential kernel."""

        # apply exponential filter
        if self.exp is None:
            pass
        else:
            raise NotImplementedError

        # remove slow changes
        try:
            del self.slow_trend
        except AttributeError:
            pass

        return signal - self.calculate_slow_trend(
            signal, self.slow_trend_window, self.min_periods_slow)

    def _calculate_baseline(self, signal):
        # first smooth with t1 rolling window
        kwargs = {'center': True, 'axis': 1}
        smooth_signal = pd.DataFrame(signal).rolling(
            window=self.t1, min_periods=int(self.min_periods_t1 * self.t1),
            **kwargs).mean()

        # then minimize with t2 rolling window
        return smooth_signal.rolling(
            window=self.t2, min_periods=int(self.min_periods_t2 * self.t2),
            **kwargs).min()


class MaxminDFOF(DFOFStrategy):
    """
    Parameters
    ----------
    window : int, optional
    sigma : float, optional
    min_periods : float (0, 1], optional

    Attributes
    ----------
    baseline : pd.DataFrame
    """

    def __init__(self, window=1800, sigma=90, min_periods=0.2):
        self.window = window
        self.sigma = sigma
        self.min_periods = min_periods

    def _filter_final_signal(self, signal):
        return signal

    def _calculate_baseline(self, signal):
        """Apply maxmin filter to signal to find baseline"""
        return filters.maxmin_filter(signal, window=self.window,
                                     sigma=self.sigma,
                                     min_periods=self.min_periods)


class Suite2pDFOF(DFOFStrategy):

    """Class for calculating dF/F that takes into account Suite2p neuropil
    estimations. Subtracting the scaled neuropil from the raw trace can be
    valuable for removing bleedthrough from nearby ROIs or other local diffuse
    signals that a low-pass baseline will obscure. However, this initial
    subtraction complicates estimating dF/F, as the resulting residual trace
    will have a baseline close to zero (or negative). Thus it is necessary to
    "add back in" the component of the low-pass baseline that was removed by
    the neuropil correction, to avoid distortions in the dF/F due to small or
    negative denominator terms.

    The unfiltered neuropil is first subtracted from the unfiltered raw trace.
    Two partial baselines are estimated then by filtering the residual raw
    trace, and the neuropil. The residual raw trace minus its partial baseline
    forms the numerator term of the dF/F. The denominator is the sum of the
    partial baselines.

    The partial baselines are calculated by applying a minimax filter to the
    respective (smoothed) traces.

    Parameters
    ----------
    window : int, optional
        Size of the rolling window for min/max/smoothing operations, in frames.
        Defaults to 600
    sigma : float
        Standard deviation of the gaussian smoothing kernel applied prior to
        min/max filtering, in frames. Defaults to 10.
    min_periods : float (0, 1], optional
        Fraction  of frames in each window that must be non-NaN in order to
        return a non-NaN baseline. Defaults to 0.2.
    constant_denominator : bool, optional
        Whether to divide the dF result by a constant denominator when
        calculating dF/F, taken as the median of the total baseline across
        all time points. This is helpful to avoid artifactually increasing
        signal amplitudes over time due to a monotonically decaying baseline,
        for example during photobleaching. Defaults to False

    Attributes
    ----------
    baseline : tuple
        Contains 3 DataFrames containing the total baseline ([0]),
        signal residual baseline ([1]), and neuropil baseline ([2]).
    """

    def __init__(self, window=600, sigma=10, min_periods=0.2,
                 constant_denominator=False):
        self.window = window
        self.sigma = sigma
        self.min_periods = min_periods
        self.constant_denominator = constant_denominator

    def _load_signals(self, experiment, channel, label):
        sig_dict = {
            'signal': experiment.signals(signal_type='raw', channel=channel,
                                         label=label, max_frame=None),
            'npil': experiment.signals(signal_type='npil', channel=channel,
                                       label=label, max_frame=None)
        }
        return sig_dict

    def _filter_final_signal(self, signal):
        """No final filtering"""
        return signal

    def _calculate_baseline(self, signal, npil):
        """doc string """

        sig_residual = signal - npil

        baseline_sig = filters.maxmin_filter(
            sig_residual, window=self.window, sigma=self.sigma,
            min_periods=self.min_periods)
        baseline_npil = filters.maxmin_filter(
            npil, window=self.window, sigma=self.sigma,
            min_periods=self.min_periods)

        baseline_total = baseline_sig + baseline_npil
        if self.constant_denominator:
            baseline_total = np.median(baseline_total)
        self.baseline = (baseline_total, baseline_sig, baseline_npil)

        return self.baseline

    def calculate(self, signal, npil):
        """Calculate dF/F

        Parameters
        ----------
        signal : array-like (n_rois, n_timepoints)
        npil : array-like (n_rois, n_timepoints)

        Returns
        -------
        dfof : pd.DataFrame
        """

        if len(signal.shape) != 2:
            raise ValueError("Input signal must be 2D")

        try:
            del self.baseline
        except AttributeError:
            pass

        base = self.calculate_baseline(signal, npil)
        return self._filter_final_signal(((signal - npil) - base[1]) / base[0])



import pandas as pd


@pd.api.extensions.register_dataframe_accessor("level")
@pd.api.extensions.register_series_accessor("level")
class LevelAccessor:
    """Syntactic sugar for working with multi-indexed dataframes.

    Examples
    ---------
    First import
    >>> from lab3.misc import level_accessor

    To get a single level from the index:
    >>> df.level['expt']

    To get multiple levels from the index
    >>> df.level[['expt', 'mouse']]

    To get rows where levels take certain values
    >>> df.level(mouse='mickey', expt='22001')

    To select rows in a more complex way
    >>> df.level('expt == 2200 | expt == 2201',
                 'duration < 600', mouse='mickey')

    For acceptable syntax refer to:
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.query.html
    """

    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def __getitem__(self, item):
        try:
            return self._obj.index.get_level_values(item)
        except KeyError:
            levels = self._obj.reset_index()[item]
            return pd.MultiIndex.from_frame(levels)

    def __call__(self, *query, **kwargs):
        query = list(query)
        for k, v in kwargs.items():
            try:
                # This may not work for namespacing reasons...
                assert (v[0] == "@") or (v[0] == v[-1] == '`')
                query.append(f'`{k}` == {v}')
            except (AssertionError, TypeError):
                query.append(f'`{k}` == "{v}"')

        try:
            return self._obj.query('(' + ') & ('.join(query) + ')' )
        except AttributeError:
            df = self._obj.to_frame()
            return df.query('(' + ') & ('.join(query) + ')')[self._obj.name]



from __future__ import annotations

import numpy as np
from typing import TYPE_CHECKING

from .base import Event
from .group import BehaviorEventGroup

if TYPE_CHECKING:
    from lab3.experiment import BehaviorExperiment

class LapEvent(Event):
    constructor_ids = Event.constructor_ids + ['lap_num'] 

    def __init__(self, lap_num, **event_kwargs):
        self.lap_num = lap_num
        super().__init__(**event_kwargs)

    def label(self):
        return self.lap_num

class LapGroup(BehaviorEventGroup):
    @classmethod
    def from_expt(cls, expt : BehaviorExperiment):
        try:
            lap_starts = expt.format_behavior_data()['lap']
        except KeyError:
            print(f"Handling missing laps for trial {expt.trial_id}! "
                     f"(Maybe look into why this happens?)")
            lap_starts, = np.where(np.diff(expt.format_behavior_data()['treadmillPosition']) < -0.75)
            lap_starts = np.concatenate([np.array([0]), lap_starts])
        nframes = expt.imaging_dataset.num_frames
        laps = [LapEvent(lap_num=i, onset=lap_starts[i], offset=lap_starts[i+1]) 
                    for i in range(len(lap_starts)-1)]
        laps += [LapEvent(lap_num=len(lap_starts)-1, onset=lap_starts[-1], offset=nframes)]
        return cls(laps)




import numpy as np
import scipy.sparse as sparse

from collections import OrderedDict
from sima.ROI import ROI, ROIList

class CNMFData(object):
    def __init__(self, Y=None, Ypath=None, A=None, C=None, b=None, f=None, S=None, template=None, dims=None, patch_structure=None):
        self.dims = dims
        self.A = AROIs(matrix=A, dims=dims)
        self.C = C
        self.b = b
        self.f = f 
        self.S = S
        self.template=template
        self.Y = Y
        self.Ypath = Ypath

        if patch_structure is not None:
            self.slices = self.calculatePatches(patch_structure)

    def calculatePatches(self, patch_structure):
        slices = {}

        if type(patch_structure[0]) is int:
            assert len(patch_structure) == len(self.template.shape)
            self.patch_tuples = it.product(*[range(k) for k in patch_structure])
            grid = [np.linspace(0, n, k+1, dtype=int) for n,k in zip(self.shape, patch_structure)]

            for comb in self.patch_tuples:
                sls = []
                tups = []
                for dim,dimidx in enumerate(comb):
                    beg_end = (grid[dim][dimidx], grid[dim][dimidx+1])
                    tup.append(beg_end)
                    sl.append(slice(*beg_end))
                slices[tuple(tup)] = tuple(sl)
        elif type(structure[0]) is tuple:
            for tup in structure:
                slices[tup] = tuple([slice(t) for t in tup])

        return slices


class AROIs(object):
    def __init__(self, roilist=None, matrix=None, tensor=None, dims=None, hold_data=False,
                 label=None, rids=None, labels=None, tags=None):
        self._matrix = None
        self._tensor = None
        self._roilist = None
        self.label = label
                
        self.hold_data = hold_data

        # sima-specific parameters
        self.labels = labels
        self.tags = tags
        
        if matrix is not None:
            if dims is None:
                raise ValueError("Matrix form is ambiguous without dims")
            else:
                self.dim0, self.dim1, self.dim2 = dims
            
            _,self.nrois = matrix.shape
            
            if sparse.issparse(matrix):
                self._matrix = matrix
            else:
                self._matrix = sparse.csc_matrix(matrix)
                  
        elif tensor is not None:
            self.nrois,self.dim0,self.dim1,self.dim2 = tensor.shape
            self._tensor = tensor
            self._matrix = self.getMatrix()
            
        elif roilist is not None:
            self.nrois = len(roilist)
            self.dim0,self.dim1,self.dim2 = roilist[0].im_shape
            
            if rids is None:
                rids = map(lambda x: x.id, roilist)
            
            self._roilist = roilist
            self._matrix = self.getMatrix()
            
        else:
            raise ValueError("Must initialize with one of 'roilist', 'matrix', or 'tensor'")
        
        if rids is not None and len(rids) == len(self):
            self.rids = OrderedDict([(rid,i) for i,rid in enumerate(rids)])
        else:
            self.rids = OrderedDict([(i,i) for i in range(len(self))])
        
                    
    @property
    def matrix(self):
        return self.getMatrix()

    @property
    def dense(self):
        return self.getMatrix(sparsified=False)
    
    @property
    def tensor(self):
        return self.getTensor()
    
    @property
    def roilist(self):
        return self.getROIList()
    
    @property
    def dims(self):
        return self.dim0, self.dim1, self.dim2

    @property
    def shape(self):
        return self.dims

    def getMatrix(self, sparsified=True):
        if self._matrix is not None:
            return self._matrix if sparsified else np.asarray(self._matrix.todense())
        
        elif self._tensor is not None:
            nrois,d0,d1,d2 = self._tensor.shape
            
            assert (d0 == self.dim0) & (d2 == self.dim2) & (d1 == self.dim1)
            
            reshaped_array = self._tensor.transpose((1, 3, 2, 0))
            reshaped_array = reshaped_array.reshape((self.dim0*self.dim1*self.dim2, nrois),
                                                   order="C")
            
            self._matrix = sparse.csc_matrix(reshaped_array)
            
            if not self.hold_data:
                self._tensor = None
        
            return self._matrix if sparsified else reshaped_array
        
        elif self._roilist is not None:
            masks = [[np.asarray(plane.todense()) for plane in roi.mask] for roi in self._roilist]
            masks = [np.stack(mask,axis=0) for mask in masks]

            self.rids = {roi.id:i for i,roi in enumerate(self._roilist)}
            self.tags = {roi.id:roi.tags for roi in self._roilist}
            self.labels = {roi.id:roi.label for roi in self._roilist}
            
            # I am evil >:D
            self._tensor = np.stack(masks, axis=0)
            matrix = self.getMatrix(sparsified=sparsified)
            
            if not self.hold_data:
                self._roilist = None
                
            return matrix
    
    def getTensor(self, transpose=None):
        if self._tensor is not None:
            return self._tensor
        
        elif self._matrix is not None:
            npx, nrois = self._matrix.shape
            
            assert npx == self.dim0*self.dim1*self.dim2
            
            array = self.getMatrix(sparsified=False)
            
            reshaped_array = array.reshape((self.dim0,self.dim2,self.dim1,nrois),
                                          order="C")
            # New order: roi x z x X x y
            reshaped_array = reshaped_array.transpose((3,0,2,1))
            
            if transpose is not None:
                reshaped_array = reshaped_array.transpose(transpose)
            
            if self.hold_data:
                self._tensor = reshaped_array
        
            return reshaped_array
        
        else:
            raise AssertionError("I don't know how to make a tensor. This should never happen")
        
    def getROIList(self,expt=None,as_bool=False,bin_threshold=0.):
        if self._roilist is not None:
            return self._roilist
        elif self._matrix is not None:
            roilist = []
                        
            As = self.getMatrix(sparsified=False)

            if expt is not None:
                starttime = expt.get("startTime")
            else:
                starttime=None
            
            for rid, idx in self.rids.items():
                A = As[:,idx]
                dim0,dim1,dim2 = self.dims
                if rid == idx:
                    roi_id = "{}_{}".format(starttime,
                                        rid)
                else:
                    roi_id = rid
                mask = A.reshape((dim0,dim2,dim1)).transpose((0,2,1))
                if as_bool:
                    mask = mask > bin_threshold

                if hasattr(self, "labels") and self.labels is not None and rid in self.labels:
                    label = self.labels[rid]
                else:
                    label = self.label

                if hasattr(self, "tags") and self.tags is not None and rid in self.tags:
                    tags = self.tags[rid]
                else:
                    tags = []

                roi = ROI(mask=mask, id=roi_id, label=label, tags=tags)
                roilist.append(roi)
            roilist = ROIList(roilist)
           # roilist = sorted(roilist, key=lambda x: x.id)

            unique, cts = np.unique(self.rids, return_counts=True)
            if not np.all(cts==1):
                for rid,num in zip(unique[cts!=1],cts[cts!=1]):
                    print("WARNING: {} occurs {} times".format(rid,num))
                    
            if self.hold_data:
                self._roilist = roilist
            # For now; eventually just save
            return roilist
        else:
            raise AssertionError("I don't know how to make an roilist. This should never happen")
        
    def keys(self):
        return self.rids.keys()

    def to_patches(self, patches, nonzero=False):
        for patch in patches:

            rois_patch = self.tensor[(slice(None),) + patch]
            rids = self.rids.keys()

            if nonzero is True:
                nonzero_roi_idxes, = np.nonzero(rois_patch.sum(axis=tuple(np.arange(1,len(self.dims)+1))))
                rids = [rids[i] for i in nonzero_roi_idxes]
                rois_patch = rois_patch[nonzero_roi_idxes]

            yield AROIs(tensor=rois_patch, rids=rids)

    def __iter__(self):
        for i, rid in enumerate(self.rids.keys()):
            yield AROIs(matrix=self.matrix[:, i], 
                         dims=self.dims, rids=[rid])

    def __len__(self):
        return self.nrois
    
    def __getitem__(self, key):
        if type(key) is slice or (type(key) is tuple and type(key[0]) is slice):
            return AROIs(matrix=self.matrix[key])

        elif type(key) is tuple:
            k,fmt = key
            
            idx = self.rids[k]
            
            if fmt == 'tensor':
                return self.tensor[idx,:,:,:]
            elif fmt == 'ROI':
                return self.roilist[idx]
            elif fmt == 'dense':
                return self.getMatrix(sparsified=False)[:,idx]
            else:
                return self[k]
            
        elif type(key) is list:
            return [self[k] for k in key]
                
        else:        
            idx = self.rids[key]
            return AROIs(matrix=self.matrix[:,idx], dims=self.dims)



"""Convenience classes for converting old data to the new signal storage
format in LAB3.

Note these methods don't actually work... what is `unpickle_old`?!"""

import os
import pandas as pd
import pickle as pkl

from lab3.experiment.base import ImagingMixin, ImagingOnlyExperiment


class OldImagingMixin(ImagingMixin):

    def signals(self, signal_type='raw', channel='Ch2', label=None,
                max_frame=None):
        channel_idx = self.imaging_dataset._resolve_channel(channel)
        if signal_type == 'raw':
            signals = self._load_old_signals(channel=channel_idx,
                                             label=label)
        elif signal_type == 'dfof':
            signals = self._load_old_dfof(channel=channel_idx,
                                          label=label)
        else:
            raise NotImplementedError(
                f'Unsupported signal type `{signal_type}`')

        if max_frame is not None:
            signals = signals.iloc[:, :(max_frame - 1)]

        return signals

    def write_newstyle(self, signal_type='raw', channel='Ch2',
                       label=None, **kwargs):
        signals = self.signals(signal_type=signal_type, channel=channel,
                               label=label, **kwargs)
        signal_key = f"/{channel}/{label}/{signal_type}"
        with self.signals_file('a') as signal_file:
            signal_file.put(signal_key, signals)

    def write_all(self, channel='Ch2'):
        channel_idx = self.imaging_dataset._resolve_channel(channel)
        signals_file = os.path.join(self.sima_path, f'signals_{channel_idx}.pkl')
        dfof_file = os.path.join(self.sima_path, f'dFoF_{channel_idx}.pkl')

        raw_data = unpickle_old(signals_file)
        dfof_data = unpickle_old(dfof_file)

        for label, data in raw_data.items():
            try:
                roi_ids = pd.Index([r['label'] for r in data['rois']], name='roi_label')
                raw = data['raw'][0]
                signal_key = f"/{channel}/{label}/raw"
                signals = pd.DataFrame(raw, index=roi_ids)
                with self.signals_file('a') as signal_file:
                    print(f"Writing {signal_key}...")
                    signal_file.put(signal_key, signals)

                # Is there dfof? If so, write
                dfof = dfof_data[label]['traces'][...,0]
                signal_key = f"/{channel}/{label}/dfof"
                with self.signals_file('a') as signal_file:
                    print(f"Writing {signal_key}...")
                    signal_file.put(signal_key, pd.DataFrame(dfof, index=roi_ids))
            except Exception as exc:
                print(f"Failed to write `{signal_key}` because of {exc}")

    def _load_old_signals(self, channel=0, label=None):
        signals_file = os.path.join(self.sima_path, f'signals_{channel}.pkl')
        data = unpickle_old(signals_file)[label]
        roi_ids = pd.Index([r['label'] for r in data['rois']], name='roi_label')
        raw = data['raw'][0]
        return pd.DataFrame(raw, index=roi_ids)

    def _load_old_dfof(self, channel=0, label=None):
        signals_file = os.path.join(self.sima_path, f'signals_{channel}.pkl')
        data = unpickle_old(signals_file)[label]
        roi_ids = pd.Index([r['label'] for r in data['rois']], name='roi_label')

        dfof_file = os.path.join(self.sima_path, f'dFoF_{channel}.pkl')
        data = unpickle_old(dfof_file)[label]
        dfof = data['traces'][..., 0]
        return pd.DataFrame(dfof, index=roi_ids)


class OldImagingExperiment(OldImagingMixin, ImagingOnlyExperiment):
    """Useful for expeditious conversion to newstyle.
    Probably *not* so good for doing extensive analysis on its own

    TODO : stick an example here on how to use objects to convert old data
    """
    pass

def unpickle_old(filename):
    with open(filename, 'rb') as f:
        data = pkl.load(f, encoding='latin1')
    return data




from __future__ import annotations
import numpy as np
import pandas as pd

from typing import Dict, TYPE_CHECKING

from lab3.analysis.base import SignalAnalysis
from lab3.core import Analysis
from lab3.core.filters import IndexIdentity, ColumnIdentity

from lab3.core.helpers import save_load

if TYPE_CHECKING:
    from lab3.experiment import ImagingMixin

class _SingleEventPSTH(SignalAnalysis):

    name = 'PSTH'
    constructor_ids = ['pre', 'post', 'dt', 'agg']

    def __init__(self, pre: float = 3., post: float = 3., dt: float = 0.03, agg: str = 'sum'):
        self.pre = pre
        self.post = post
        self.dt = dt
        self.agg = agg

    def to_columns(self):
        return pd.TimedeltaIndex(np.arange(-self.pre, self.post+self.dt, self.dt), 
                                    name=self.name, unit='s')

    
    def apply_to(self, event, signals):
        psth = self._calculate_single_event(event.trigger, 
                                            signals, 
                                            pre=self.pre, 
                                            post=self.post) 
        return self._align_time(psth)

    def _calculate_single_event(self, trigger, signals, pre, post):

        # WHY CAN'T THIS BE NAMED self.calculate?
        # Ideally this should live in self.calculate, but for some 
        # reason super() gets lost if I do that

        start = pd.Timedelta(seconds=trigger-pre)
        stop = pd.Timedelta(seconds=trigger+post)
        psth = signals.loc[:, start:stop]
        psth.columns -= pd.Timedelta(seconds=trigger)

        return psth 


    def _align_time(self, psth):
        if self.agg == 'sum':
            new_psths = psth.T.resample(f"{self.dt}S").sum()
        elif self.agg == 'mean':
            new_psths = psth.T.resample(f"{self.dt}S").mean()
        else:
            # EXPERIMENTAL!
            new_psths = psth.T.resample(f"{self.dt}S").agg(self.agg)

        # TODO: Why doesn't this work
        new_psths = new_psths.interpolate('time', limit=10)
        new_psths = new_psths.reindex(index=self.columns, 
                                      method='nearest') # level=self.name
        new_psths = new_psths.T

        return new_psths


class PSTH(_SingleEventPSTH):
    """Calculates a PSTH for a single experiment based on a list of events or an event_spec. 
    For each neuron, we compute its activity in a pre/post window around event triggers. 

    Initialization parameters
    ----------
    pre : float
        size of pre-window, in seconds
    post : float
        size of post-window, in seconds
    dt : float
        temporal resolution, in seconds. If there is data collected at multiple
        sampling rates, it will all be resampled to dt. 

    Examples
    --------
    >>> # Compute the running PSTH in a peri-3 second window
    >>> strategy = PSTH(pre=3, post=3, dt=0.1)
    >>> signal_spec = {"signal_type": "dfof", "label": "suite2p"}
    >>> event_spec = {"signal_type": "behavior", "event_type": "running", "as_time_events": True}
    >>> psths = strategy.apply_to(
    >>>     expt, signal_spec=signal_spec, events=)
    """
    @save_load
    def apply_to(self, expt : ImagingMixin, events, signal_spec, roi_filter=IndexIdentity, 
                    time_filter=ColumnIdentity, trial_filter=IndexIdentity):
        # signals should just do this
        signals = expt.signals(**signal_spec, as_time_signals=True)
        signals = signals.loc[roi_filter.bind_to(expt)]
        signals.loc[:, ~time_filter.bind_to(expt)] = np.nan
        #signals.columns = expt.imaging_times
        signals.columns = pd.TimedeltaIndex(signals.columns, unit='s')#expt.imaging_times(), unit='s')
        # signals should just do this
        signals.index = signals.index.set_names('roi_label')

        # Instead of passing an event_list directly, can pass an
        # event "spec" to read from experiment
        events = self._load_events(expt, events, trial_filter=trial_filter)

        return self.calculate(signals, events)

    def calculate(self, signals, events, **kwargs):
        # Subclassing makes calling this method very pretty
        return events.apply(super(), signals=signals)

class Responsiveness(PSTH):
    """Calculates mean activity pre and post an event trigger. 
    """
    name = 'Responsiveness'
    def to_columns(self):
        return pd.Index(['pre', 'post'], name=self.name)

    def __init__(self, pre: float = 3., 
                    post: float = 3., dt: float = 0.03):
        super().__init__(pre=pre, post=post, dt=dt)
        self.psth_strategy = PSTH(pre=pre, post=post, dt=dt)

    def calculate(self, signals, events):
        psth = self.psth_strategy.calculate(signals, events)

        if psth is None:
            return None

        trigger = pd.Timedelta(seconds=0)
        pre = psth.loc[:, :trigger].mean(axis=1)
        post = psth.loc[:, trigger:].mean(axis=1)

        return pd.DataFrame({'pre': pre, 'post': post}, columns=self.columns)

    



from __future__ import annotations

import inspect
import numpy as np 
import pandas as pd

from abc import abstractmethod
from typing import TYPE_CHECKING

from lab3.core import Analysis
from lab3.core.filters import IndexIdentity, ColumnIdentity

if TYPE_CHECKING:
    from lab3.experiment import ImagingMixin

class BaseSignalAnalysis:

    """Base class for all signal analysis strategies."""

    @classmethod
    def _get_param_names(cls):
        init_signature = inspect.signature(cls.__init__)
        parameters = [p for p in init_signature.parameters.values()
                      if p.name != 'self' and p.kind != p.VAR_KEYWORD]

        return sorted([p.name for p in parameters])

    def get_params(self):
        out = dict()
        for key in self._get_param_names():
            try:
                value = getattr(self, key)
            except AttributeError:
                value = None
            out[key] = value
        return out

    def set_params(self):
        raise NotImplementedError

    def __repr__(self, line_max=80):
        """TODO : Make this pretty"""
        repr_ = type(self).__name__
        repr_ += '('
        base_length = len(repr_)
        curr_line_length = len(repr_)
        for key, value in self.get_params().items():
            param_string = key + '=' + str(value) + ', '
            if (curr_line_length + len(param_string)) <= line_max:
                repr_ += param_string
                curr_line_length += len(param_string)
            else:
                repr_ += '\n' + ' ' * base_length
                repr_ += param_string
                curr_line_length = base_length + len(param_string)
        repr_ = repr_[:-2] + ')'
        return repr_

    def __str__(self):
        return self.__repr__()

class SignalAnalysis(Analysis):
    @abstractmethod
    def apply_to(self, expt : ImagingMixin, signal_spec, roi_filter=None, time_filter=None, **kwargs):
        signals = self._load_signals(expt, signal_spec)
        signals = signals.loc[roi_filter.apply(expt), time_filter.apply(expt)]

        result = self._calculate(signals, **kwargs)
        """
        DO STUFF HERE
        """
        return result

    def _calculate(self, signals, **kwargs):
        """
        DO STUFF HERE
        """
        result.columns = self.columns # suggested, not mandated
        raise NotImplementedError

    def _load_signals(self, expt, signal_spec, roi_filter=IndexIdentity(),
                    time_filter=ColumnIdentity()):
        """Helper method for loading signals.
        """
        signals = expt.signals(**signal_spec)
        signals.index.name = 'roi_label'
        signals = signals.loc[roi_filter.bind_to(expt)]
        signals.loc[:, ~time_filter.bind_to(expt)] = np.nan

        return signals


    def _load_events(self, expt, events, bool_array=False,
                     inverse=False, trial_filter=IndexIdentity):
        """Helper method for loading events.
        """
        try:
            events['signal_type']
            events = expt.events(**events)
            events = events.filter_trials(trial_filter.bind_to(expt))
        except TypeError:
            pass

        if bool_array:
            event_array = events.bool_array(
                            expt.imaging_dataset.num_frames)
            if inverse:
                event_array = ~event_array

            return event_array

        return events




import os

import numpy as np
import pandas as pd
import rhd

from lab3.signal import SignalFile


# label, channel, signal_type
def overwrite_lfp(base_key, overwrite, signal_file):
    if base_key in signal_file:
        assert overwrite, f"Signals already exist for {base_key}. " \
            + "Set overwrite=True to re-import.\nCurrent file structure:" \
            + f" \n {signal_file.info()}"

        print(f"Deleting previous import at {base_key}")
        signal_file.remove(base_key)


def trim_to_imaging(rhd_data):
    try:
        rhd_data['trimmed']
        return rhd_data
    except KeyError:
        expt_start = np.where(rhd_data['board_adc_data'][1] > 1)[0][0]

        # Signals
        rhd_data['amplifier_data'] = rhd_data['amplifier_data'][:, expt_start:]
        rhd_data['board_adc_data'] = rhd_data['board_adc_data'][:, expt_start:]

        # Times
        rhd_data['t_amplifier'] = rhd_data['t_amplifier'][expt_start:]
        rhd_data['t_board_adc'] = rhd_data['t_board_adc'][expt_start:]

        # Re-zero
        rhd_data['t_amplifier'] -= rhd_data['t_amplifier'][0]
        rhd_data['t_board_adc'] -= rhd_data['t_board_adc'][0]

        rhd_data['trimmed'] = True

        return rhd_data


def import_intan_to_lfp_signals_file(
        intan_path, lfp_signals_path, label='LFP', preprocess_func=None,
        amplifier_channels=None, adc_channels=None, overwrite=False):
    """Import rhd to an lfp_signals.h5

    Parameters
    ----------
    intan_path : str
        Path to .rhd file
    lfp_signals_path : str
        Directory where LFP data is stored
    preprocess_func : callable, optional
        Preprocessing on output of rhd.read_data. This should take a
        dictionary as input and return the same dictionary as output, with
        the preprocessing you want performed.
    label : str, optional
        Label to be applied to the 'amplifier_data' channels (ignored if
        'amplifier_channels' is specified)
    amplifier_channels : iterable, optional
        Labels to interpret amplifier channels as
    adc_channels : iterable, optional
        Labels to interpret ADC channels as
    overwrite : bool
        Whether to overwrite existing data
    """

    data = rhd.read_data(intan_path)

    if preprocess_func is not None:
        data = preprocess_func(data)

    with SignalFile(os.path.join(lfp_signals_path,
                                 'lfp_signals.h5')) as signal_file:
        print("Storing...")
        if amplifier_channels is None:
            overwrite_lfp(label, overwrite, signal_file)

            print(f"Deleting previous import at {base_key}")
            signal_file.remove(base_key)

            raw = pd.DataFrame(data['amplifier_data'],
                               columns=data['t_amplifier'],
                               index=np.arange(len(data['amplifier_data'])))

            signal_file.put(f"{label}/raw", raw)
        else:
            raw = pd.DataFrame(data['amplifier_data'],
                               columns=data['t_amplifier'],
                               index=amplifier_channels)
            for channel_name in raw.index:
                overwrite_lfp(channel_name, overwrite, signal_file)
                signal_file.put(f"{channel_name}/raw",
                                raw.loc[channel_name].to_frame(channel_name))

        if adc_channels is None:
            overwrite_lfp("ADC", overwrite, signal_file)
            adc = pd.DataFrame(data['board_adc_data'],
                               columns=data['t_board_adc'],
                               index=np.arange(len(data['board_adc_data'])))
            signal_file.put(f"ADC/raw", adc)
        else:
            adc = pd.DataFrame(data['board_adc_data'],
                               columns=data['t_board_adc'],
                               index=adc_channels)
            for channel_name in adc.index:
                overwrite_lfp(channel_name, overwrite, signal_file)
                signal_file.put(f"{channel_name}/raw",
                                adc.loc[channel_name].to_frame(channel_name))
        # signal_file["metadata"] = data['frequency_parameters']
        print("Done!")



import numpy as np
from scipy.signal import resample


def calc_cutoff(length, fs, max_trunc=None):
    """DSP trick to speed up resampling

    Parameters
    ----------
    length : int
        Length of original signal
    fs : float
        Sampling rate of original signal
    max_trunc : float, optional
        Maximum allowable truncation (1 / [fs units])
    """
    if max_trunc is None:
        return None
    else:
        max_trunc_samples = fs * max_trunc
        exponent = int(np.log2(max_trunc_samples))
        assert exponent > 0
        cutoff = int(length / (2 ** exponent)) * 2 ** exponent
        return cutoff


def fast_resample(x, num, t=None, axis=0, fs=None, max_truncation=None,
                  truncate_output=False, **kwargs):
    """Uses a DSP trick to speed up resampling

    Parameters
    ----------
    x : array
        Signal to resample
    num : int
        Number of samples in output
    t : array, optional
        Times to sample
    axis : int, optional
        Defaults to 0
    fs : float, optional
        Sampling rate of original signal
    max_truncation : float, optional
        Maximum allowable truncation (1 / [fs units])
    truncate_output : bool, optional
        Apply same trick to the output (NOTE: this results in at most 2x
        max_truncation)

    Notes
    -----
    This only supports downsampling for now. Either t or fs must be specified!
    Sampling rate should be about uniform
    """

    if (t is None) and (fs is None):
        raise ValueError("One of 't' or 'fs' must be specified!")

    length = x.shape[axis]
    if num > length:
        raise NotImplementedError("Only downsampling is supported for now")

    if t is not None:
        fs = 1 / np.mean(np.diff(t))

    cutoff = calc_cutoff(length, fs, max_trunc=max_truncation)

    truncated_x = np.moveaxis(np.moveaxis(x, axis, 0)[:cutoff], 0, axis)

    if truncate_output:
        resample_factor = num / length
        resampled_fs = fs * resample_factor
        output_length = calc_cutoff(num, resampled_fs,
                                    max_trunc=max_truncation)
        return resample(truncated_x, output_length, t=t, axis=axis, **kwargs)
    else:
        return resample(truncated_x, num, axis=axis, t=t, **kwargs)



# General Dependencies
import os
import numpy as np
import h5py
import warnings

from numba import jit

import sima
from sima import ImagingDataset
from sima.sequence import _Sequence_HDF5

warnings.warn(
"""
Did you link mkl libraries with e.g.
    export LD_PRELOAD=/opt/intel/mkl/lib/intel64/libmkl_def.so:/opt/intel/mkl/lib/intel64/libmkl_avx2.so:/opt/intel/mkl/lib/intel64/libmkl_core.so:/opt/intel/mkl/lib/intel64/libmkl_intel_lp64.so:/opt/intel/mkl/lib/intel64/libmkl_intel_thread.so:/opt/intel/lib/intel64_lin/libiomp5.so
"""
)

"""
CHECK IF RUNNING IN NOTEBOOK TO PRE-EMPT ERROR
"""
try:
    ip = get_ipython()
    if ip.has_trait('kernel'):
        warnings.warn(
"""
###############################################################################
You seem to be running Trefide in a Jupyter notebook. Trefide doesn't play nice 
with Jupyter for reasons I don't understand. Known issues include:
    - Importing fails unless the compiled shared objects (.so) are in the 
    same directory as the .ipynb (doesn't happen with plain command line)
    - Parallelization doesn't work
    - Sometimes you get SegFaults
Proceed with caution... -Z
###############################################################################
"""
            )
except:
    pass

# Preprocessing Dependencies
from trefide.utils import psd_noise_estimate

# PMD Model Dependencies
from trefide.pmd import batch_decompose,\
                        batch_recompose,\
                        overlapping_batch_decompose,\
                        overlapping_batch_recompose,\
                        determine_thresholds
from trefide.reformat import overlapping_component_reformat, weighted_component_reformat

def parse_mmapname(name):
    name = os.path.basename(name)

    parts = name.split('_')
    assert parts[-1] == '.mmap'
    data = {
        'T': int(parts[-2]),
        'order': parts[-4],
        'z': int(parts[-6]),
        'x': int(parts[-8]),
        'y': int(parts[-10]),
        'name': '_'.join(parts[:-10])
    }

    data['shape'] = (data['T'], data['z'], data['x'], data['y'])

    return data 

def trim_sequence_to_block(seq, block_height, block_width, tsub):
    """Sequence size must evenly divide block_height and width
    """
    fov_height = (seq.shape[2]//block_height)*block_height
    fov_width = (seq.shape[3]//block_width)*block_width
    num_frames = (seq.shape[0] // tsub)*tsub


    btrim = (seq.shape[2] - fov_height)//2    
    ltrim = (seq.shape[3] - fov_width)//2
    ttrim = seq.shape[0] - num_frames
    
    mov = np.array(seq).transpose((2,3,0,1,4))[btrim:btrim+fov_height, ltrim:ltrim+fov_width, ttrim:, 0, 0]
    mov = np.ascontiguousarray(np.nan_to_num(mov))

    return mov

def load_mmap_trimmed(mmapname, block_height, block_width, t_sub):

    mmap_info = parse_mmapname(mmapname)
    mmap = np.memmap(mmapname, shape=mmap_info['shape'], 
                        order='F', mode='r') # F BECAUSE TRANSPOSE!

    fov_height = (mmap.shape[2]//block_height)*block_height
    fov_width = (mmap.shape[3]//block_width)*block_width
    num_frames = (mmap.shape[0] // t_sub)*t_sub


    btrim = (mmap.shape[2] - fov_height)//2    
    ltrim = (mmap.shape[3] - fov_width)//2
    ttrim = mmap.shape[0] - num_frames

    real_mmap = mmap.transpose((3,2,0,1))[btrim:btrim+fov_height, ltrim:ltrim+fov_width, ttrim:, 0]
    print(real_mmap.flags)
    return real_mmap

def load_mmap_raw(mmapname):
    mmap_info = parse_mmapname(mmapname)
    mmap = np.memmap(mmapname, shape=mmap_info['shape'], dtype=np.double,
                        order='F', mode='r+')
    return mmap.transpose((3,2,0,1))[...,0]

def denoise(mmapname, sima_compatible=False, max_components=50, max_iters_main=10, max_iters_init=40, 
                d_sub=2, t_sub=2, consec_failures=3, tol=5e-3, block_height=40, 
                block_width=40, overlapping=True, enable_temporal_denoiser=True, 
                enable_spatial_denoiser=True):

    """Run Trefide denoising on the imaging dataset. It should ordinarily not be 
    necessary to change the Trefide parameters from defaults.

    Parameters
    ----------
    mmapname : str
        Path to memmap containing data to denoise.
    sima_compatible : bool, optional
        Whether to reconstitute the full format (i.e., all frames). Default False.
    overlapping : bool, optional
        Whether to overlap blocks the video is split into. Default True.
    block_height : int, optional 
        Frame height will be trimmed to a multiple of this number. Default 40.
    block_width: int, optional
        Frame width will be trimmed to a multiple of this number. Default 40.
    d_sub : int, optional
        Factor by which to downsample spatial dimensions. Default 2. 
    t_sub : int, optional 
        Factor by which to downsample time. Default 2. 
    max_components : int, optional
        Cap on the maximum number of components. Default 50. 
    max_iters_main : int, optional
        Cap on maximum number of iterations in main method. Default 10. 
    max_iters_init : int, optional
        Cap on maximum number of initialization iterations. Default 40. 
    consec_failures : int, optional
        Number of consecutive failures to tolerate. Default 3. 
    tol : float, optional
        Tolerance. Default 5e-3. 
    enable_temporal_denoiser : bool, optional
        Whether to use temporal denoising. Default True. 
    enable_spatial_denoiser : bool, optional
        Whether to use spatial denoising. Default True.
    """

    #mov = load_mmap_trimmed(mmapname, block_height, block_width, t_sub)
    mov = load_mmap_raw(mmapname) #TESTING ONLY
    # mov = trim_sequence_to_block(seq, block_height, block_width)
    fov_height, fov_width, num_frames = mov.shape 

    # Calculate denoising thresholds
    print("Calculating thresholds...")
    spatial_thresh, temporal_thresh = determine_thresholds((fov_height, fov_width, min(num_frames, 3000)), # don't spend forever doing this
                                                   (block_height, block_width),
                                                   consec_failures, max_iters_main, 
                                                   max_iters_init, tol, 
                                                   d_sub, t_sub, 5, False,
                                                   enable_temporal_denoiser,
                                                   enable_spatial_denoiser)

    print(spatial_thresh, temporal_thresh)

    if not overlapping:    # Blockwise Parallel, Single Tiling
        print("Decomposing blockwise in parallel using single tiling...")
        spatial_components,\
        temporal_components,\
        block_ranks,\
        block_indices = batch_decompose(fov_height, fov_width, num_frames,
                                        mov, block_height, block_width,
                                        spatial_thresh, temporal_thresh,
                                        max_components, consec_failures,
                                        max_iters_main, max_iters_init, tol,
                                        d_sub, t_sub,
                                        enable_temporal_denoiser, enable_spatial_denoiser)

        U, V = weighted_component_reformat(spatial_components, temporal_components, 
                                            block_ranks, block_indices, 1)
    else:    # Blockwise Parallel, 4x Overlapping Tiling
        print("Decomposing blockwise in parallel using 4x overlapping tiling...")
        spatial_components,\
        temporal_components,\
        block_ranks,\
        block_indices,\
        block_weights = overlapping_batch_decompose(fov_height, fov_width, num_frames,
                                                    mov, block_height, block_width,
                                                    spatial_thresh, temporal_thresh,
                                                    max_components, consec_failures,
                                                    max_iters_main, max_iters_init, tol,
                                                    d_sub, t_sub,
                                                    enable_temporal_denoiser, enable_spatial_denoiser)
        
        print("Reformatting to compressed format...")
        U, V = overlapping_component_reformat(fov_height, fov_width, num_frames,
                                  block_height, block_width,
                                  spatial_components,
                                  temporal_components,
                                  block_ranks,
                                  block_indices,
                                  block_weights)
    
    # if sima_compatible:
    #     print("Rebuilding movie...")


    #     # if not overlapping:  # Single Tiling (No need for reweighting)
    #     #     mov_denoised = np.asarray(batch_recompose(spatial_components,
    #     #                               temporal_components,
    #     #                               block_ranks,
    #     #                               block_indices))
    #     # else:   # Overlapping Tilings With Reweighting
    #     #     mov_denoised = np.asarray(overlapping_batch_recompose(fov_height, fov_width, num_frames,
    #     #                               block_height, block_width,
    #     #                               spatial_components,
    #     #                               temporal_components,
    #     #                               block_ranks,
    #     #                               block_indices,
    #     #                               block_weights)) 
    # else:
    mov_denoised = None


    return U, V, mov_denoised


@jit
def rebuild_movie(U, V, out=None):
    if out is None:
        return U.dot(V)
    else:
        # This is optimized for OOC operations!
        assert out.shape == U.shape[:-1] + V.shape[-1:]
        assert U.shape[-1] == V.shape[0]
        assert not np.any(out)
        k, T = V.shape
        for i in range(k):
#            for t in range(T):
            out += U[...,i:i+1].dot(V[i:i+1]) # ,t



import pandas as pd 
import copy
from abc import abstractmethod
from functools import reduce

from . import Item, Analysis, Group
from lab3.core.helpers import conjunction, disjunction
from lab3.misc import level_accessor

class Filter(Item):
    """Generic filter. Filters a dataframe, usually a signals dataframe (rois or timepoints)
    based on some criterion. Works by implicitly building an AST on the stack.
    
    Usage should be something like
    >>> roi_filter = ROIFilterSubclass(...)
    >>> desired_rois = roi_filter.bind_to(expt)
    >>> time_filter = TimeFilterSubclass(...)
    >>> desired_timepoints = time_filter.bind_to(expt)
    >>> signals.loc[desired_rois, desired_timepoints]
    
    This should be subclassed rather than used directly. 

    IMPORTANT! You MUST use `f1 & f2`, and NEVER `f1 and f2` (likewise for `or`). 
    The `and`/`or` binary operators cannot be overridden in Python, thus 
    `f1 and f2` results in UNDEFINED BEHAVIOR! -Z.

    Read more:
    https://www.python.org/dev/peps/pep-0335/
    """
    inverse = False

    def bind_to(self, item):
        """Defines the indices to select, and binds to an Item, returning a
        _BoundFilter instance
        """
        raise NotImplementedError 

    def __or__(self, other):
        return FilterGroup([self, other], disjunction=True)        
        
    def __add__(self, other):
        return self | other

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            return self | other
    
    def __sub__(self, other):
        return self | -other
    
    def __and__(self, other):
        return FilterGroup([self, other], conjunction=True)
    
    def __mul__(self, other):
        return self & other

    def __xor__(self, other):
        # When would this ever be necessary?
        return (self | other) & ~(self & other)
    
    def __invert__(self):
        newself = copy.deepcopy(self)
        newself.inverse = not self.inverse
        return newself
    
    def __neg__(self):
        return ~self
    
    def __repr__(self):
        if self.inverse:
            return "~" + self.name
        else:
            return self.name

    def _repr_html_(self):
        return repr(self)
    
class _BoundFilter(Filter):
    """Object bound to data and used to slice a DataFrame.
    Created by a Filter, *Not* instantiated directly

    TODO: this^^^ is not an accurate description of what this class is/does
    """
    constructor_ids = ['index', 'columns', 'level', 'inverse', 'name']

    def __call__(self, df):
        if self.index is not None:
            return self.filter_index(df)
        elif self.columns is not None:
            return self.filter_columns(df)

    def filter_index(self, df):
        all_labels = df.index

        # Workaround to pandas bug
        # level= cannot be passed to flat Indexes
        if (self.level is not None) and (len(all_labels.names) > 1):
            # pd.Index.reindex returns some other stuff too - throw it away
            index, _ = all_labels.reindex(self.index, level=self.level)
        else:
            index = self.index

        if self.inverse:
            return all_labels.difference(index)
        else:
            return index

    def filter_columns(self, df):
        all_labels = df.columns
        if self.inverse:
            return all_labels.difference(self.columns)
        else:
            return self.columns

    def __or__(self, other):
        return _BoundFilterGroup([self, other], disjunction=True)        
        
    def __and__(self, other):
        return _BoundFilterGroup([self, other], conjunction=True)

    def __repr__(self):
        return super().__repr__() + "[BOUND]"

class FilterGroup(Filter, Group):
    """Group of filters

    """
    constructor_ids = ['name', 'conjunction', 'disjunction']
    def __init__(self, iterable_of_filters, conjunction=False, 
                    disjunction=False, name='FilterGroup'):
        super().__init__(iterable_of_filters, name=name)
        if not conjunction ^ disjunction:
            raise ValueError("EITHER conjunction or disjunction must be True!")
        self.conjunction = conjunction
        self.disjunction = disjunction
    
    def bind_to(self, *args, **kwargs):
        return _BoundFilterGroup([filt.bind_to(*args, **kwargs) 
                                    for filt in self], 
                                    conjunction=self.conjunction, 
                                    disjunction=self.disjunction)

    def __invert__(self):
        # Distribute using DeMorgan's Laws
        return self.__class__([~filt for filt in self], conjunction=not self.conjunction, 
                                disjunction=not self.disjunction, name=self.name)
        
    def __repr__(self):
        INV = "~"*self.inverse
        
        if self.conjunction:
            return f"{INV}({' ∧ '.join([repr(filt) for filt in self])})"
        elif self.disjunction:
            return f"{INV}({' ∨ '.join([repr(filt) for filt in self])})"

class _BoundFilterGroup(FilterGroup, _BoundFilter):
    def __call__(self, df):
        if self.conjunction:
            return conjunction(self)(df)
        elif self.disjunction:
            return disjunction(self)(df)
        
class ROIFilter(Filter):
    def bind(self, index, level='roi_label'):
        return _BoundFilter(index=index, columns=None, level=level, 
                            inverse=self.inverse, name=self.name)

class TimeFilter(Filter):
    def bind(self, columns):
        return _BoundFilter(index=None, columns=columns, level=None, 
            inverse=self.inverse, name=self.name)

class _IndexIdentity(ROIFilter):
    name = 'IndexIdentity'
    inverse = True
    def bind_to(self, expt):
        return self.bind([], level=None)

    def __call__(self):
        return self

class _ColumnIdentity(TimeFilter):
    name = 'ColumnIdentity'
    inverse = True 
    def bind_to(self, expt):
        return self.bind([])

    def __call__(self):
        return self

IndexIdentity = _IndexIdentity()
ColumnIdentity = _ColumnIdentity()




import sima
import shutil
import os
import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.collections import PolyCollection

from lab3.experiment import ImagingOnlyExperiment, utils
from lab3.extraction.s2p.batch import BatchSuite2pImagingDataset

class PostHocIdentificationDataset(BatchSuite2pImagingDataset):
    def __init__(self, functional_sessions, savedir, static_session=None, static_tiff=None):
        child_dirs = []
        
        if isinstance(functional_sessions, list):
            child_dirs.extend(functional_sessions)
        elif isinstance(functional_sessions, str):
            child_dirs.append(functional_sessions)
        elif functional_sessions is None:
            # Load from existing directory
            super().__init__(None, savedir)
            self.static_savedir = self.child_dirs[-1]
            return
        else:
            raise ValueError(f"Invalid sessions {functional_sessions}")
            
        if static_session is not None:
            self.static_savedir = static_session
            child_dirs.append(static_session)
        elif static_tiff is not None:
            tiff_seq = sima.Sequence.create("TIFF", path=static_tiff)
            
            frame = tiff_seq._get_frame(0)
            
            ndarray_seq = sima.Sequence.create("ndarray", frame.reshape((1,) + frame.shape))
            
            self.static_savedir = static_tiff.replace('.tif', '.sima')
            
            if os.path.exists(self.static_savedir):
                shutil.rmtree(self.static_savedir)
            
            static_ds = sima.ImagingDataset([ndarray_seq], 
                                            savedir=self.static_savedir, 
                                            channel_names=['Ch2']) # intentionally fictitious
            child_dirs.append(self.static_savedir)
        else:
            raise ValueError("Either `static_session` or `static_tiff` must be specified!")

        if os.path.exists(savedir):
            shutil.rmtree(savedir)
        
        super().__init__(child_dirs=child_dirs, savedir=savedir)
        
    def import_red_rois(self, threshold=None, threshold_pct=99., green_label='PYR', red_label='red', 
                        current_label='suite2p', 
                        plot_summary=False, **kwargs):
        """
        
        Run this *after* doing `import_results_to_children`
        
        """
        static_expt = ImagingOnlyExperiment(self.static_savedir)
        signals = static_expt.signals(label=current_label, signal_type='raw').mean(axis=1)

        if not threshold:
            threshold = np.percentile(static_expt.imaging_dataset.time_averages, threshold_pct)

        new_label_rois = signals.loc[signals > threshold].index
        
        for sima_path in self.child_dirs:
            expt = ImagingOnlyExperiment(sima_path)
            utils.split_signals(expt, new_label=red_label, 
                                new_label_rois=new_label_rois, 
                                current_label=current_label, 
                                **kwargs)

            utils.split_signals(expt, new_label=green_label, 
                    new_label_rois=new_label_rois, 
                    current_label=current_label, 
                    inverse=True,
                    **kwargs)
            
        if plot_summary:
            self.plot_summary(green_label=green_label, red_label=red_label)
        
    def plot_summary(self, green_label=None, red_label=None):
        static_expt = ImagingOnlyExperiment(self.static_savedir)
        img = static_expt.imaging_dataset.time_averages[0,...,0]
        
        pct1 = np.nanpercentile(img.flatten(), 1)
        pct99 = np.nanpercentile(img.flatten(), 98)
        
        rois = static_expt.roi_list(label=green_label)
        
        if red_label is not None:
            red_rois = static_expt.roi_list(label=red_label)
        else:
            red_rois = []
            
        with PdfPages(os.path.join(self.savedir, 'extraction_summary.pdf')) as pdf:
            fig, ax = plt.subplots(figsize=(10,10))
            ax.imshow(img, cmap='gray', vmin=pct1, vmax=pct99, aspect='equal')
            
            all_roi_verts = []
            red_verts = []
            
            for r in rois:
                all_roi_verts.append(r.coords[0][:,:2])
                
            for r in red_rois:
                red_verts.append(r.coords[0][:,:2])
            
            collL = PolyCollection(all_roi_verts,alpha=0.5)
            collR = PolyCollection(red_verts,alpha=0.5, color='r')
            ax.add_collection(collL)
            ax.add_collection(collR)



import os
import cv2
import glob
import psutil
import numpy as np
import scipy
import sys
import time

from scipy.ndimage.filters import gaussian_filter
from skimage.external.tifffile import TiffFile

try:
    cv2.setNumThreads(0)
except:
    pass

import caiman as cm
from caiman.motion_correction import MotionCorrect, tile_and_correct, motion_correction_piecewise

import caiman.source_extraction.cnmf as cnmf
from caiman.components_evaluation import evaluate_components, estimate_components_quality_auto

def motion_correct(memmap_name, dview, is3D=False,
                  max_shifts=(6, 6), strides=(48, 48), overlaps=(24, 24), 
                  num_frames_split=100,
                  max_deviation_rigid=3, 
                  pw_second_pass=False,
                  pw_rigid=False,
                  shifts_opencv=True, 
                  nonneg_movie=True,
                  border_nan='copy', **normcorre_kwds):

    indices = tuple([slice(None) for _ in max_shifts])

    mc = MotionCorrect([memmap_name], dview=dview, max_shifts=max_shifts,
                  strides=strides, overlaps=overlaps, pw_rigid=pw_rigid,
                  max_deviation_rigid=max_deviation_rigid, 
                  shifts_opencv=shifts_opencv, nonneg_movie=True,
                  border_nan=border_nan, is3D=is3D, indices=indices, **normcorre_kwds)

    mc.motion_correct(save_movie=True)

    if pw_second_pass:
        print("Running second pass nonrigid motion correction...")
        mc.pw_rigid = True
        mc.template = mc.mmap_file
        mc.motion_correct(save_movie=True, template=mc.total_template_rig)

    return mc

def extract(mmapname, dview, n_processes=8, cnmf_second_pass=True, rf=15, stride=10, K=12, 
            gSig=(2, 2), merge_thresh = 0.8, p=0, fr=10, decay_time=1., gnb=2,
            min_SNR=3, rval_thr=0.7, use_cnn=False, **cnmf_kwds):
        # amounpl.it of overlap between the patches in pixels
        # number of neurons expected per patch
        # expected half size of neurons
        # merging threshold, max correlation allowed
         # half-size of the patches in pixels. rf=25, patches are 50x50
       # order of the autoregressive system)
    # approx final rate  (after eventual downsampling )
     # length of typical transient in seconds 
    # CNN classifier is designed for 2d (real) data
    # accept components with that peak-SNR or higher
     # accept components with space correlation threshold or higher


    Yr, dims, T = cm.load_memmap(mmapname)

    # Does this load the data into RAM???
    images = np.reshape(Yr.T, [T] + list(dims), order='C')    # reshape data in Python format (T x X x Y x Z)

    cnm = cnmf.CNMF(n_processes=n_processes, dview=dview, k=K, gSig=gSig, 
                stride=stride, gnb=gnb,  #rf=rf, # this parameter seems problematic 
                merge_thresh=merge_thresh, p=p, only_init_patch=True, **cnmf_kwds)

    cnm = cnm.fit(images)

    print(('Number of components:' + str(cnm.estimates.A.shape[-1])))

    # Evaluate components
    cnm.params.change_params(params_dict={'fr': fr,
                                          'decay_time': decay_time,
                                          'min_SNR': min_SNR,
                                          'rval_thr': rval_thr,
                                          'use_cnn': use_cnn})

    cnm.estimates.evaluate_components(images, cnm.params, dview=dview)
    print(('Keeping ' + str(len(cnm.estimates.idx_components)) +
            ' and discarding  ' + str(len(cnm.estimates.idx_components_bad))))
    
    if cnmf_second_pass:
        print("Second pass CNMF (seeded with first pass)...")
        cnm.params.set('temporal', {'p': p})
        cnm2 = cnm.refit(images)
        
        print("Computing DF/F")
        cnm2.estimates.detrend_df_f()

        return cnm2
    else:
        print("Computing DF/F")
        cnm.estimates.detrend_df_f()
        return cnm  





import os
import shutil
import itertools as it
import multiprocessing

import numpy as np
from suite2p.run_s2p import run_s2p
try:
    from suite2p.io.utils import init_ops
except ModuleNotFoundError:
    from suite2p.io.utils import init_ops
from sima.sequence import _fill_gaps

from lab3.misc.progressbar import ProgressBar
from lab3.misc.pool_helpers import (bind_locked_get_frame, pool_helper,
                                    slice_sequence)
from lab3.misc.sima_sequences import _NumPyFuncSequence


sequence_store = 0
lock = 0


def dump_to_binary(ds, signal_channel='Ch2', static_channel=None,
                   fill_gaps=False):
    """Converts SIMA imaging dataset to binary file, in the format expected by
    Suite2p. Along the way, additional 'ops' dictionaries are created for
    each imaging plane and saved to disk in the suite2p directory.

    If the dataset has multiple sequences, they will be concatenated in the
    binary file as if they were one continuous recording. For this reason,
    all sequences should be of compatible frame shape (zyxc). This allows batch
    processing of datasets from the same FOV, but special consideration should
    be given for motion correction. If the individual datasets have already
    been motion corrected, you must align the sequences and crop them
    accordingly when constructing the batch imaging dataset object.
    This step is unnecessary if using Suite2p's motion correction, as the
    frames of all sequences will be aligned to a common reference.

    Parameters
    ----------
    ds : Suite2pStrategy (sima.imaging.ImagingDataset)
        A Suite2p strategy instance (e.g. Suite2pImagingDataset), initialized
        from the sima folder with the desired imaging dataset. All sequences in
        the imaging dataset will be concatenated and written to the binary
        file.
    signal_channel : str or int
        Channel with dynamic signal (e.g. GCaMP). Defaults to 'Ch2'.
    static_channel : str or int
        Channel with static marker (e.g. tdTomato). If passed, the static
        channel is also dumped to a binary file. Defaults to None.
    fill_gaps : bool
        Fill in NaNed pixels with data from adjacent frames. If False, any
        NaNed pixels will appear as zeros in the binary file. Defaults to
        False.

    Returns
    -------
    ops_list : list of dicts
        List of ops dictionaries (one for each imaging plane). If there is only
        one plane, or collapse_z=True, it will be a list of one dict.

    Notes
    -----
    The format of the binary file and the addtiional parameters added to the
    ops dictionaries has not changed from 0.5.5 to the latest versions as far
    as we are aware, so this function should generalize.
    """

    # create list of channels to dump to binary
    channel_list = [ds._resolve_channel(signal_channel)]
    if static_channel is not None:
        channel_list.append(ds._resolve_channel(static_channel))
        reg_file_chan2, meanImg_chan2 = [], []

    Ly, Lx = ds.frame_shape[1:3]

    reg_file, meanImg = [], []
    ops_list = init_ops({**ds.ops, **ds.db})

    for pidx, ops in enumerate(ops_list):

        # open binary for writing
        reg_file.append(open(ops['reg_file'], 'wb'))
        meanImg.append(np.zeros((Ly, Lx)))

        if static_channel is not None:
            reg_file_chan2.append(open(ops['reg_file_chan2'], 'wb'))
            meanImg_chan2.append(np.zeros((Ly, Lx)))

    # note that if not fill_gaps, nans will be set to zero in the binary
    save_frames = [_fill_gaps(iter(seq), iter(seq))
                   if fill_gaps else iter(seq)
                   for seq in ds.sequences]

    # iterate over sequence and dump to binaries
    # TODO clean this up
    frame_counter = 0
    for sidx, seq in enumerate(save_frames):
        p = ProgressBar(ds.sequences[sidx].shape[0])

        for fidx, frame in enumerate(seq):
            frame_counter += 1
            p.update(fidx)

            for pidx, plane in enumerate(frame):
                for ch_idx, channel in enumerate(channel_list):
                    if ch_idx == 0:
                        reg_file[pidx].write(bytearray(
                            plane[:, :, channel].astype('uint16')))
                        meanImg[pidx] += plane[:, :, channel]
                    else:
                        reg_file_chan2[pidx].write(bytearray(
                            plane[:, :, channel].astype('uint16')))
                        meanImg_chan2[pidx] += plane[:, :, channel]
        p.end()

    # save plane options, close binary files
    for pidx, ops in enumerate(ops_list):
        ops['Ly'] = Ly
        ops['Lx'] = Lx
        ops['nframes'] = frame_counter
        ops['meanImg'] = meanImg[pidx] / frame_counter
        if not ops['do_registration']:
            ops['refImg'] = ops['meanImg']
        reg_file[pidx].close()

        if static_channel is not None:
            ops['meanImg_chan2'] = meanImg_chan2[pidx] / frame_counter
            reg_file_chan2[pidx].close()

        np.save(ops['ops_path'], ops)

    # save ops_list
    np.save(ds.ops_list_path, ops_list)
    return ops_list


def _write_data(sidx, paths, fill_gaps, signal_channel, static_channel):
    """ Method for performing parallel dump of imaging data into binary files.

    Parameters
    ----------
    sidx : int
        the index of the sequence in the global sequence store to write.
    paths : list of str
        the folder path to store temporary data in for each of the reg_files.
    fill_gaps : bool
        Fill in NaNed pixels with data from adjacent frames. If False, any
        NaNed pixels will appear as zeros in the binary file. Defaults to
        False.
    signal_channel : int
        Channel index with dynamic signal (e.g. GCaMP).
    static_channel : int
        Channel index with static marker (e.g. tdTomato). If passed, the static
        channel is also dumped to a binary file.

    Returns
    -------
    tmp_file_paths : list of str
        list of temporary files, one per plane of imaging data,  storing the
        imaging data converted to the binary format stored.
    meanImg : numpy.array
        the sum of all of the imaging frames exported, used in calculating the
        mean image for ths dataset.
    """
    global sequence_store
    global lock

    seq = sequence_store[sidx]
    tmp_filepaths = [os.path.join(path, 'data_%s.bin' % sidx)
                     for path in paths]
    tmp_files = list(map(lambda p: open(p, 'wb'), tmp_filepaths))
    if static_channel is not None:
        tmp_filepaths_chan2 = [os.path.join(path, 'data_chan2_%s.bin' % sidx)
                               for path in paths]
        tmp_files_chan2 = list(map(lambda p: open(p, 'wb'),
                                   tmp_filepaths_chan2))

    meanImg = np.zeros(seq.shape[1:])

    # note that if not fill_gaps, nans will be set to zero in the binary
    seq_iter = _fill_gaps(iter(seq), iter(seq)) if fill_gaps else iter(seq)

    p = ProgressBar(seq.shape[0])
    for fidx, frame in enumerate(seq_iter):
        p.update(fidx)
        meanImg += frame
        for pidx, plane in enumerate(frame):
            tmp_files[pidx].write(
                bytearray(plane[..., signal_channel].astype('int16')))
            if static_channel is not None:
                tmp_files_chan2[pidx].write(
                    bytearray(plane[..., static_channel].astype('int16')))

    list(map(lambda f: f.close(), tmp_files))

    if static_channel is not None:
        list(map(lambda f: f.close(), tmp_files_chan2))
        tmp_filepaths = [tmp_filepaths, tmp_filepaths_chan2]
    else:
        tmp_filepaths = [tmp_filepaths]

    p.end()
    return (tmp_filepaths, meanImg)


def _append_files(tmp_files, reg_files, sidx):
    """ Concatenates the data in binary files. Usefull when performing parallel
    dump to binary.

    Parameters
    ----------
    tmp_files : list of list of str
        Temporary files to be combined. Files will be deleted after being
        dumped into destination file. Outer list is for each plane in the
        dataset. Inner list is each file to be combined.
    reg_files : list of str
        Resulting binary file to combine tmp_files into.
    sidx : int
        sequence index being saved. if sidx is zero then new reg_files are
        created, otherwise the data is appended to the existing reg_files
    """
    print('appending_files')
    for pidx, reg_filepath in enumerate(reg_files):
        if sidx == 0:
            shutil.move(tmp_files[pidx].pop(0), reg_filepath)
        with open(reg_filepath, 'ab') as f:
            for tmp_file in tmp_files[pidx]:
                with open(tmp_file, 'rb') as f2:
                    while True:
                        _bytes = f2.read(1024)
                        if not _bytes:
                            break
                        f.write(_bytes)
                print('removing %s' % tmp_file)
                os.remove(tmp_file)


def parallel_dump_to_binary(ds, signal_channel='Ch2', static_channel=None,
                            fill_gaps=False, n_processes=1):
    """Converts SIMA imaging dataset to binary file, in the format expected by
    Suite2p. Along the way, additional 'ops' dictionaries are created for
    each imaging plane and saved to disk in the suite2p directory.

    see the method :meth:`lab3.extraction.s2p_helpers.dump_to_binary`

    Parameters
    ----------
    ds : Suite2pStrategy (sima.imaging.ImagingDataset)
        A Suite2p strategy instance (e.g. Suite2pImagingDataset), initialized
        from the sima folder with the desired imaging dataset. All sequences in
        the imaging dataset will be concatenated and written to the binary
        file.
    signal_channel : str or int
        Channel with dynamic signal (e.g. GCaMP). Defaults to 'Ch2'.
    static_channel : str or int
        Channel with static marker (e.g. tdTomato). If passed, the static
        channel is also dumped to a binary file. Defaults to None.
    fill_gaps : bool
        Fill in NaNed pixels with data from adjacent frames. If False, any
        NaNed pixels will appear as zeros in the binary file. Defaults to
        False.

    Returns
    -------
    ops_list : list of dicts
        List of ops dictionaries (one for each imaging plane). If there is only
        one plane, or collapse_z=True, it will be a list of one dict.
    """

    global lock
    global sequence_store
    lock = multiprocessing.Lock()

    signal_channel = ds._resolve_channel(signal_channel)
    if static_channel is not None:
        static_channel = ds._resolve_channel(static_channel)

    sequences = ds.sequences
    num_planes, Ly, Lx = sequences[0].shape[1:4]
    ops_list = init_ops({**ds.ops, **ds.db})
    for pidx in range(num_planes):

        if not os.path.isdir(ops_list[pidx]['save_path']):
            os.makedirs(ops_list[pidx]['save_path'])

    # iterate over sequence and dump to binaries
    for sidx, seq in enumerate(sequences):
        if n_processes > 1:
            bind_locked_get_frame(seq, lock)
            sequence_store = list(slice_sequence(seq, n_processes=n_processes))
            pool = multiprocessing.Pool(processes=n_processes)

        else:
            pool = None
            sequence_store = [seq]

        results = pool_helper(
            pool,
            _write_data,
            range(len(sequence_store)),
            it.repeat([o['save_path'] for o in ops_list]),
            it.repeat(fill_gaps), it.repeat(signal_channel),
            it.repeat(static_channel)
        )

        if pool is not None:
            pool.close()
            pool.join()
        else:
            results = list(results)

        weights = [s.shape[0] for s in sequence_store]
        sequence_store = None
        meanImages = np.array([r[1] for r in results])
        meanImages = meanImages / np.array(weights)[:, None, None, None, None]

        tmp_files = np.array([r[0] for r in results])
        tmp_file_list = tmp_files[:, 0, :].T.tolist()
        _append_files(tmp_file_list, [o['reg_file'] for o in ops_list], sidx)
        if static_channel is not None:
            tmp_files_chan2 = tmp_files[:, 1, :].T.tolist()
            _append_files(
                tmp_files_chan2, [o['reg_file_chan2'] for o in ops_list], sidx)

    # save plane options
    for pidx, ops in enumerate(ops_list):
        ops['Ly'] = Ly
        ops['Lx'] = Lx
        ops['nframes'] = seq.shape[0]
        ops['meanImg'] = meanImages[pidx, ..., signal_channel]
        if not ops['do_registration']:
            ops['refImg'] = ops['meanImg']

        if static_channel is not None:
            ops['meanImg_chan2'] = meanImages[pidx, ..., static_channel]

        np.save(ops['ops_path'], ops)

    np.save(ds.ops_list_path, ops_list)

    return ops_list


def accept_all_rois(save_dir):
    """ Places all ROIs in a suite2p dataset in the "is cell" category.

    Parameters
    ----------
    save_dir : str
        File path to walk. All iscell.npy files contianed in this directory or
        in a subdirectory of save_dir will be edited to accept all the ROIs.
    """

    for root, dirs, files in os.walk(save_dir):
        if 'iscell.npy' in files:
            cells = np.load(os.path.join(root, 'iscell.npy'))
            cells[:, 0] = 1
            np.save(os.path.join(root, 'iscell.npy'), cells)


def find_masked_frames(seq):
    """Find indices of masked frames in sequence"""
    masked_frames = []
    current_seq = seq
    while True:
        try:
            masked_frames += current_seq._mask_dict.keys()
        except AttributeError:
            pass
        try:
            current_seq = current_seq._base
        except AttributeError:
            break

    return np.unique(masked_frames)


def extract(ds, signal_channel='Ch2', static_channel=None, fs=10,
            collapse_z=True, fill_gaps=None, remove_masked_frames=True,
            register=False, sparse_mode=True, spatial_scale=0, diameter=None,
            overwrite=False, reprocess=False, n_processes=1, ops_kws={}):
    """
    Parameters
    ----------
    signal_channel : str or int, optional
        Name or index of the dynamic channel to be extracted.
    static_channel : str or int, optional
        Name or index of the static channel. Provide this if you want
        to for example use Suite2p's red cell detection. To determine
        registration shifts using the static channel, override
        'align_by_chan' via ops_kws (see below), setting it equal to 2.
    fs : float, optional
        Frame rate (per-plane), in Hz. Defaults to 10.
    collapse_z : bool, optional
        Whether to average all planes on each frame prior to processing.
        Defaults to True.
    fill_gaps : bool, optional
        Whether to fill NaNed values with data from adjacent frames.
        Defaults to None, which will fill_gaps only if register is False.
    remove_masked_frames : bool, optional
        Skip masked frames when creating binary file. They will be
        reinserted as NaNed entries at the proper timepoints in the final
        imported signal traces.
        Not implemented.
    register : bool, optional
        Whether to use Suite2p motion correction, which is nonrigid by
        default. To use the rigid motion correction only, override
        'nonrigid' via ops_kws (see below), setting it equal to False.
    sparse_mode : bool, optional
        Whether to use Suite2p's sparse mode. Defaults to True
    spatial_scale : int {0, 1, 2, 3}, optional
        ROI scale parameter for sparse model. 0 = multiscale, 1 = 6 pixels,
        2 = 12 pixels, 3 = 24 pixels, 4 = 48 pixels. Choose the closest value.
        Defaults to 0.
    diameter : int, optional
        Expected diameter of cells, in pixels. Defaults to 15. Ignored
        if sparse_mode is True.
    overwrite : bool, optional
        If True, existing Suite2p folder is deleted and dataset is
        reprocessed from the beginning of the pipeline. Defaults to False.
    reprocess : bool, optional
        If True, rerun Suite2p analysis on existing binary file.
        This can be used for example to extract ROIs multiple times with
        different parameters, without repeating the binary conversion and
        registration each time. Reprocess and overwrite cannot both be True.
        Not tested - we may need to rewrite the ops dictionaries still.
    ops_kws : dict, optional
        Pass additional keyword-argument pairs as a dictionary to
        override other Suite2p default settings.
    """

    if collapse_z and ds.frame_shape[0] > 1:
        ds.sequences = \
            [_NumPyFuncSequence(s, 'max', axis=0) for s in ds.sequences]

    # set user parameters and create ops dictionary
    ops_kws['nchannels'] = 2 if static_channel is not None else 1
    ops_kws['fs'] = fs
    ops_kws['do_registration'] = 1 if register else 0
    ops_kws['nplanes'] = ds.frame_shape[0]
    ops_kws['sparse_mode'] = sparse_mode
    if sparse_mode:
        ops_kws['spatial_scale'] = spatial_scale
    else:
        if not diameter:
            raise ValueError("If sparse_mode is False, you must specify a "
                             + "cell diameter")
        else:
            ops_kws['diameter'] = diameter
    ds.ops = ops_kws  # any unspecified parameters take their default values

    if not reprocess:
        # convert imaging data into suite2p binary file and create/save
        # additional settings files
        if fill_gaps is None:
            fill_gaps = not register

        print("Converting imaging data to Suite2p binary file\n"
              + f"(concatenating {len(ds.sequences)} sequences)")
        if n_processes == 1:
            dump_to_binary(ds, signal_channel=signal_channel,
                           static_channel=static_channel,
                           fill_gaps=fill_gaps)
        else:
            parallel_dump_to_binary(
                ds, signal_channel=signal_channel, fill_gaps=fill_gaps,
                static_channel=static_channel, n_processes=n_processes)

    else:
        print("Re-running Suite2p on existing binary file, if possible")

        # update ops files with new parameters
        new_ops_list = ds.ops_list
        for idx, ops in enumerate(new_ops_list):
            ops = {**ops, **ds.ops, **ds.db}
            np.save(ops['save_path'], ops)
            new_ops_list[idx] = ops
        np.save(ds.ops_list_path, new_ops_list)

    # handle masked frames
    try:
        os.unlink(ds.bad_frames_path)
    except FileNotFoundError:
        pass
    if remove_masked_frames:
        bad_frames = []
        frame_counter = 0
        for seq in ds.sequences:
            bad_frames.append(find_masked_frames(seq) + frame_counter)
            frame_counter += len(seq)
        bad_frames = np.concatenate(bad_frames)

        if len(bad_frames):
            # save in npy file, which will be recognized by suite2p
            np.save(ds.bad_frames_path, bad_frames)
            print(f"Ignoring frames:\n{bad_frames}")
        else:
            print("No frames to mask")

    # run suite2p main function
    print("\n\n-------ENTERING SUITE2P-------\n\n")
    run_s2p(ds.ops, ds.db)

# def extract_with_055(self, signal_channel='Ch2', static_channel=None,
#                      collapse_z=True, fill_gaps=None,
#                      remove_masked_frames=True, register=False,
#                      x_size=10, y_size=10, overwrite=False, reprocess=False,
#                      ops_kws={}):
#     """Provides backward compatibility for extraction with Suite2p 0.5.5"""
#
#     # we can reuse the extract function I think. Just override some other
#     # arguments. we need to set sparse_mode to false and encode the x and y
#     # dimensions separately
#     raise NotImplementedError



"""A collection of base classes for representing the basic experiment types
used in the lab: combinations of behavior, imaging, and LFP recordings.
Inherit from these to extend their functionality for more paradigms, e.g.
specific behavior tasks, that lend themselves to bespoke analysis methods."""

import os
import json
import functools
import pickle as pkl
import warnings
import glob
import pandas as pd
import h5py

import numpy as np
from sima.imaging import ImagingDataset

from lab3.core import Item
from lab3.core.metadata import DefaultMetadata
from lab3.experiment import utils, database
from lab3.experiment.decorators import default_label_handling
from lab3.signal.base import SignalFile
from lab3.signal.utils import nan_zscore, nonparametric_nan_zscore

from lab3.event import BehaviorEventGroup, LFPEventGroup, EventGroup
from lab3.event.transient import ExperimentTransientGroup

# Accomodate old naming conventions
KEY_DICT = {
    'startTime': 'start_time',
    'time': 'start_time',
    'stopTime': 'stop_time',
    'tSeriesDirectory': 'tSeries_path',
    'project_name': 'experiment_group'
}


def fetch_trials(*args, **kwargs):
    """Pass keyword arguments specifying attribute and value pairs from
    the database, to get a list of trial IDs satisfying those conditions.
    Some keyword arguments are automatically renamed to match the database
    conventions. Some common attributes are listed below.

    Parameters
    ----------
    args : str, optional
        Name of trial attributes required to be present
    project_name : str, optional
        By convention, this is usually the experimenter's name
    mouse_name : str, optional
    experimentType : str, optional
    start_time : str
    stop_time : str

    Returns
    -------
    trial_ids : list

    Examples
    --------
    >>> trial_ids = fetch_trials(project_name='test', mouse_name='test')
    >>> imaged_trials = fetch_trials('tSeriesDirectory', mouse_name='test')

    See also
    --------
    lab3.experiment.database
    """

    # TODO should we continue to support this, or just strictly adopt the names
    # of the database?

    args = list(args)

    for k in KEY_DICT:
        if k in kwargs:
            kwargs[KEY_DICT[k]] = kwargs.pop(k)

        if k in args:
            args[args.index(k)] = KEY_DICT[k]

    return database.fetch_trials(*args, **kwargs)


class _BaseExperiment(Item):

    """Class establishing some basic experiment attributes. Not intended to be
    used directly"""

    constructor_ids = ['trial_id']

    def __init__(self, trial_id=None, mouse_id=''):
        self.trial_id = trial_id
        self._trial_info = {'trial_id': trial_id, 'mouse_id': mouse_id}

    @property
    def mouse_id(self):
        return self._trial_info['mouse_id']

    @property
    def experiment_type(self):
        try:
            return self._trial_info['experimentType']
        except KeyError:
            return ''

    def events(self, event_type, signal_type='imaging', channel=None,
               label=None, **kwargs):
        if signal_type == 'imaging':
            return self.imaging_events(event_type=event_type, channel=channel,
                                       label=label, **kwargs)
        elif signal_type == 'LFP':
            return self.lfp_events(event_type, channel=channel, label=label,
                                   **kwargs)
        elif signal_type == 'behavior':
            return self.behavior_events(event_type=event_type, **kwargs)

    def signals(self, signal_type='raw', channel='Ch2',
               label=None, max_frame=None, as_time_signals=False, 
               source='imaging', metadata=DefaultMetadata, **kwargs):
        if source == 'behavior':
            signals = self.behavior_signals(
                label=label, channel=channel, as_time_signals=as_time_signals,
                 **kwargs)
        elif source == 'LFP':
            signals = self.lfp_signals(
                signal_type=signal_type, channel=channel, label=label, 
                as_time_signals=as_time_signals, **kwargs)
        elif source == 'imaging':
            signals = self.imaging_signals(
                signal_type=signal_type, channel=channel, 
                as_time_signals=as_time_signals, label=label, **kwargs)
        else:
            signals = source(self, **kwargs)

        try:
            signals = metadata.bind_to(signals, expt=self)
        except Exception as exc:
            print(f"Binding metadata : {metadata} to {self} failed with {exc}")

        if max_frame is not None:
            return signals.iloc[:, :max_frame]
        else:
            return signals

class BehaviorExperiment(_BaseExperiment):

    """Base class for all experiments in the behavior database.

    Parameters
    ----------
    trial_id : int
        ID of experiment in the sql database
    """

    def __init__(self, trial_id):
        super().__init__(trial_id=trial_id)

        # get experiment attributes from the database
        # NOTE : We've dramatically simplified the loading of experiment
        # attributes from the database. There might be some growing pains for
        # things that are no longer automatically exposed that will need to be
        # reimplemented

        try:
            self._trial_info.update({
                **database.fetch_trial(trial_id),
                **database.fetch_all_trial_attrs(trial_id, parse=True)
            })
            self._trial_info_clean = self._trial_info.copy()
        except TypeError:
            raise KeyError(f"Trial ID {trial_id} does not exist")

    def identifiers(self):
        return {
            'trial_id': self.trial_id,
            'mouse_id': self.mouse_name,
            'expt_type': self.experiment_type,
            'start_time': self.start_time,
        }

    @property
    def mouse_id(self):
        return self._trial_info['mouse_id']

    @property
    def mouse_name(self):
        return database.fetch_mouse(self.mouse_id)['mouse_name']

    @property
    def start_time(self):
        return self._trial_info['start_time']

    @property
    def stop_time(self):
        return self._trial_info['stop_time']

    @property
    def filename(self):
        return self._trial_info['behavior_file']

    def __setattr__(self, item, value):
        super().__setattr__(item, value)
        if item[0:4] == '_db_':
            # Setting an attribute prefixed with `_db_` will flag it for a
            # database update if self.save is called
            # This is a work-around to avoid namespace polution. Generally
            # variables in self._trial_info should be exposed via properties
            # as needed, and protected with setters to sanitize changes
            self._trial_info[item[4:]] = value

    @property
    def archived(self):
        # Fix this! It will break as-is. What is it even for?

        try:
            return self._props['archived']
        except KeyError:
            if self.get('tSeriesDirectory') is None:
                self._props['archived'] = False
            else:
                self._props['archived'] = \
                    len(glob.glob(os.path.join(
                        self.tSeriesDirectory, '*.archive'))) > 0

            return self.archived

    def save(self, store=False):
        """Update the sql database trial information based on the keys of
        `self._trial_info`. Note you shouldn't normally read or write to this
        dictionary directly, but rather expose variables as needed by defining
        new properties. See below for example implementations.

        Parameters
        ----------
        store : bool, optional
            Whether to actually update the database. Defaults to False, which
            will just print the changes that would be made.

        Examples
        --------
        Expose a key in `self._trial_info` as a property with a corresponding
        setter method:

        >>> class DumbyExpt(BehaviorExperiment):
        >>>     @property
        >>>     def my_awesome_attribute(self):
        >>>         return self._trial_info['my_awesome_attribute']
        >>>     @my_awesome_attribute.setter
        >>>     def my_awesome_attribute(self, value):
        >>>         self._trial_info['my_awesome_attribute'] = value

        Alternatively, attributes prefixed with `_db_` will automatically
        edit the corresponding value in `self._trial_info` when they are set:

        >>>     @my_awesome_attribute.setter
        >>>     def my_awesome_attribute(self, value):
        >>>         # this updates self._trial_info['my_awesome_attribute'] !
        >>>         self._db_my_awesome_attribute = value

        Set the attribute value and update the database:

        >>> expt = DumbyExpt(1)
        >>> expt.my_awesome_attribute = 'foo'
        >>> expt.save(store=True)
        """

        # compare to the clean dict created during init to find edits
        updates = {k: v for k, v in self._trial_info.items() if k not in
                   self._trial_info_clean.keys()
                   or v != self._trial_info_clean[k]}

        update_trial = False
        trial_args = ['behavior_file', 'mouse_name', 'start_time', 'stop_time',
                      'experiment_group']

        if not store:
            print(f'changes to {self.trial_id}: {updates}')
        else:
            print(f'saving changes to {self.trial_id}: {updates}')
            for key, value in updates.items():
                if key == 'trial_id':
                    raise Exception('trial_id cannot be changed')
                elif key in trial_args and key != 'mouse_name':
                    update_trial = True
                else:
                    if value is None:
                        database.delete_trial_attr(self.trial_id, key)
                    else:
                        if isinstance(value, dict) or isinstance(value, list):
                            value = json.dumps(value)
                        database.update_trial_attr(self.trial_id, key, value)

            self._trial_info_clean = self._trial_info.copy()

        if update_trial:
            database.update_trial(*[self.get(k) for k in trial_args],
                                  trial_id=self.trial_id)

    def delete(self):
        database.delete_trial(self.trial_id)
        self._trial_info = {}

    @classmethod
    def create(cls, tdml_path, repickle=False, overwrite=False):
        """Parse and pickle a tdml behavior file and create a database entry,
        returning the corresponding BehaviorExperiment object. If the
        experiment already exists in the database, simply return the
        BehaviorExperiment object, unless overwrite is passed.

        Parameters
        ----------
        tdml_path : str
            Path to tdml behavior file from which to create the experiment
        repickle : bool, optional
            Whether to reparse and pickle tdml file, if pickle file already
            exists.
        overwrite : bool, optional
            Delete and recreate the database entry for this behavior file.
            Defaults to False

        Returns
        -------
        Instance of BehaviorExperiment
        """

        assert (tdml_path.endswith('.tdml') and os.path.exists(tdml_path)), \
            "Not a valid tdml file"
        pkl_path = tdml_path.replace('.tdml', '.pkl')

        # make sure tdml file is pickled, re-pickling if desired
        if os.path.exists(pkl_path) and not repickle:
            pass
        else:
            from lab3.misc.tdml_pickler import pickle_file
            pickle_file(tdml_path)

        # (re)create database entry, or just get the trial_id if not overwrite
        trial_id = database.add_experiment_to_database(
            pkl_path, overwrite=overwrite)

        return cls(trial_id)

    def repickle_tdml(self):
        """Re-parse and pickle the tdml file associated with this experiment"""
        from lab3.misc.tdml_pickler import pickle_file
        pickle_file(self.filename.replace('.pkl', '.tdml'))

    def pair_imaging_data(self, sima_path, force_pairing=False, store=True):
        """Pair experiment data with imaging data from a sima folder,
        updating the database and returning the corresponding
        ImagingExperiment object.

        Parameters
        ----------
        sima_path : str
            Path to sima folder for imaging data
        force_pairing : bool, optional
            If this experiment is already paired with a sima folder, whether to
            overwrite pairing. Default is False
        store : bool, optional
            Whether to store the newly paired sima_path in the database.
            Defaults to True.
            """
        return ImagingExperiment(self.trial_id, sima_path,
                                 force_pairing=force_pairing,
                                 store=store)

    @property
    def behavior_data(self):
        """Get unformatted behavior dictionary from pkl file."""
        with open(self.filename.replace('.tdml', '.pkl'), 'rb') as f:
            data = pkl.load(f, encoding='latin1')
        return data

    @functools.lru_cache(maxsize=5)
    def format_behavior_data(self, sampling_interval=0.1, discard_pre=0,
                             discard_post=np.inf, sigma=0.1):
        """Format behavior dictionary. Interval variables are converted to
        boolean vectors that denote activity at each sampling period.
        Continuous variables are re-sampled at each discrete sampling period.

        Parameters
        ----------
        sampling_interval : float, optional
            Sampling rate for discretizing time. Variables expressed as
            intervals (i.e. vectors of start and stop times) are converted to
            binary vectors that are True for frames inside the intervals.
            Continuous variables are resampled at the corresponding time
            points.
        discard_pre : {'first_lap', float}, optional
            If 'first_lap', truncate data before the first complete lap. If
            float, truncate data occuring before this time. Default behavior
            returns all data points
        discard_post : float, optional
            Truncate data occuring after this time. Default behavior returns
            all data points
        sigma : float, optional
            Standard deviation of gaussian smoothing kernel for smoothing
            velocity variable, in seconds. Defaults to 0.1

        Returns
        -------
        beh_dict : dict
            Behavior data dictionary with variables expressed in discrete time
            bins

        See also
        --------
        experiment.utils.discretize_behavior
        """

        return utils.discretize_behavior(self.behavior_data,
                                         sampling_interval=sampling_interval,
                                         discard_pre=discard_pre,
                                         discard_post=discard_post,
                                         sigma=sigma)

    def velocity(self, sigma=0.1, **beh_kws):
        """Get velocity trace.

        Parameters
        ----------
        sigma : float, optional
            Standard deviation of the gaussian smoothing kernel to apply to the
            velocity trace, in seconds. Defaults to 0.1.

        Keyword arguments are passed directly to format_behavior_data, and can
        be used to modify sampling rate and trimming.
        """

        return self.format_behavior_data(sigma=sigma, **beh_kws)['velocity']

    def discrete_position(self, num_bins=100, **beh_kws):
        """Get discrete position bin for each behavior sample

        Parameters
        ----------
        num_bins : int, optional
            Divide track into this number of discrete position bins. Defaults
            to 100

        Keyword arguments are passed directly to format_behavior_data, and can
        be used to modify sampling rate and trimming.
        """

        continuous_position = self.format_behavior_data(
            **beh_kws)['treadmillPosition']
        return np.floor(continuous_position * num_bins).astype('int')

    def laps(self, **beh_kws):
        """Get lap number for each behavior sample

        Keyword arguments are passed directly to format_behavior_data, and can
        be used to modify sampling rate and trimming.
        """

        return self.format_behavior_data(**beh_kws)['lap_bin']

    def behavior_signals(self, label, signal_type=None, channel=None, 
                         as_time_signals=False, sampling_interval=0.1, 
                         **kwargs):
        behavior_signals = self.format_behavior_data(
            sampling_interval=sampling_interval, **kwargs)
        signals = behavior_signals[label]
        signals = pd.DataFrame(signals, columns=[label]).T

        if as_time_signals:
            signals.columns = \
                signals.columns * behavior_signals['sampling_interval']
                
        return signals

    def behavior_events(self, event_type, **kwargs):
        return BehaviorEventGroup.load(self, event_type, **kwargs)

    def signals(self, *args, source='behavior', **kwargs):
        return super().signals(*args, source=source, **kwargs)        

class ImagingMixin:

    """Mixin for 2-photon imaging experiments. Provides methods for accessing
    and manipulating extracted imaging signals."""

    @property
    def frame_rate(self):
        return 1 / self.frame_period

    @property
    def frame_period(self):
        return self.imaging_parameters['frame_period']

    @property
    def imaging_system(self):
        return self.imaging_parameters['imaging_system']

    @property
    def imaging_parameters(self):
        """Returns a dictionary of imaging parameters from the h5 file"""
        try:
            return self._params
        except (KeyError, AttributeError):
            params_file = os.path.join(self.sima_path, "imaging_parameters.json") 
            try:
                with open(params_file) as f:
                    self._params = json.load(f)
                return self._params
            except FileNotFoundError:
                from .utils import NumpyEncoder
                h5 = h5py.File(self.imaging_dataset.sequences[0]._path, 'r')
                self._params = {k: v for k, v in h5['/imaging'].attrs.items()}

                with open(params_file, "w") as f:
                    json.dump(self._params, f, cls=NumpyEncoder)

                return self._params

    @property
    def imaging_dataset(self):
        """Return the sima ImagingDataset associated with this experiment"""
        return ImagingDataset.load(self.sima_path)

    def imaging_times(self, max_frame='trim_to_behavior'):

        if max_frame == 'trim_to_behavior':
            # figure out length of behavior data in imaging frames
            beh_dict = self.format_behavior_data()
            max_frame = beh_dict['recordingDuration'] / self.frame_period
        elif max_frame is None:
            max_frame = self.imaging_dataset.num_frames  #self.signals().shape[1] (shouldn't this be the same?!)
        return np.arange(max_frame)*self.frame_period

    @property
    def sima_path(self):
        """Return the sima directory paired to this experiment"""
        return self._db_sima_path

    @sima_path.setter
    def sima_path(self, path):
        path = os.path.normpath(path)
        assert path.endswith('.sima') and os.path.isdir(path), \
            "Not a valid sima path"
        self._db_sima_path = path

    @property
    def signals_path(self):
        return os.path.join(self.sima_path, "signals.h5")

    @property
    def imaging_events_path(self):
        return os.path.join(self.sima_path, "events.h5")

    @property
    def suite2p_imaging_dataset(self, wrapper_class=None):
        """Cast SIMA dataset as a Suite2p strategy.

        Parameters
        ----------
        wrapper_class : Suite2pStrategy, optional
            Class with which to instantiate the Suite2p imaging dataset.
            Defaults to the standard
            :class:`~lab3.extraction.s2p.Suite2pImagingDataset` class.
            Override to use your own custom subclass.
        """

        if wrapper_class is None:
            from ..extraction.s2p import Suite2pImagingDataset
            wrapper_class = Suite2pImagingDataset

        return wrapper_class.load(self.sima_path)

    def signals_file(self, mode='r', **kwargs):
        """Returns the HDFStore containing extracted and processed signals,
        read only"""
        return SignalFile(self.signals_path, mode=mode,  **kwargs)

    @default_label_handling
    def imaging_signals(self, signal_type='raw', channel='Ch2', 
                        as_time_signals=False, label=None, max_frame=None, 
                        z_score=False):
        """Retrieve signals from signals.h5.

        Parameters
        ----------
        signal_type : str {'raw', 'npil', 'dfof', 'spikes', or custom}
            Name of signal type, defaults to 'raw'
        channel : {str, int}, optional
            Channel name or index in the imaging dataset
        label : str, optional
            ROI list label in the imaging dataset. If None, the most recently
            accessed label in the imaging dataset will be used
        max_frame : int, optional
            Only return the first max_frame frames. Defaults to None (all
            frames are returned)
        z_score : bool or str {'nonparametric'}, optional
            Return the z-scored signal. If 'nonparametric', will use median and
            IQR in lieu of mean and stdev
        Returns
        -------
        signals : pd.DataFrame
        """

        with self.signals_file() as signal_file:
            signals = signal_file.get(f"/{channel}/{label}/{signal_type}")

        if z_score is True:
            signals = signals.apply(nan_zscore, axis=1)
        elif z_score == 'nonparametric':
            signals = signals.apply(nonparametric_nan_zscore, axis=1)

        if as_time_signals:
            signals.columns = self.imaging_times(max_frame=max_frame)

        return signals

    def signals(self, *args, source='imaging', **kwargs):
        return super().signals(*args, source=source, **kwargs)

    def raw_fluorescence(self, **kwargs):
        return self.signals(signal_type='raw', **kwargs)

    def dfof(self, **kwargs):
        return self.signals(signal_type='dfof', **kwargs)

    def spikes(self, **kwargs):
        return self.signals(signal_type='spikes', **kwargs)

    def check_overwrite(self, store_key, overwrite):
        with self.signals_file() as signals_file:
            key_exists = store_key in signals_file
            assert not key_exists or overwrite, \
                f"Key {store_key} already exists in signal file. Set " \
                + "overwrite=True to recalculate"

    def imaging_events(self, event_type, channel='Ch2', label=None, **kwargs):
        """Retrieve imaging events from events.h5"""
        if 'transient' in event_type:
            return ExperimentTransientGroup.load(self, event_type=event_type, 
                                            channel=channel, label=label, **kwargs)
        else:
            # This is probably incorrect for imaging-events since it does differentiate by cell
            return EventGroup.load(self, event_type=event_type,
                               channel=channel, label=label, **kwargs)

    @default_label_handling
    def calculate_signals(self, strategy, channel='Ch2', label=None,
                          to_type='misc', overwrite=False, **kwargs):
        """Run a signal processing analysis.

        Parameters
        ----------
        strategy : lab3.core.Automorphism
            Instance of a valid signal analysis.
        channel : {str, int}, optional
            Channel label or index in imaging dataset
        label : str, optional
            ROI label in imaging dataset. If None, the most recently accessed
            label in the imaging dataset will be used
        to_type : str, optional
            Name of signal type, for storing in signals.h5. The store key
            for the results will become channel/label/to_type.
        overwrite : bool, optional
            Whether to overwrite signal key if it exists already. Defaults to
            False
        **kwargs
            Additional keyword arguments are passed directly to
            strategy.apply_to()

        Returns
        -------
        pd.DataFrame
        """

        store_key = f"/{channel}/{label}/{to_type}"
        self.check_overwrite(store_key, overwrite)

        # replace this with a check for a more specific SignalAnalysis class
        from lab3.core import Automorphism
        assert isinstance(strategy, Automorphism), \
            "strategy must be an instance of an Automorphism class!"

        s = self.apply(strategy, channel=channel, label=label, **kwargs)

        with self.signals_file(mode='a') as signals_file:
            signals_file.put(store_key, s)

        return s

    def infer_spikes(self, strategy, **kwargs):
        """Run spike inferencing on signals.

        Parameters
        ----------
        strategy : SpikesStrategy
            Instance of a SpikesStrategy. See lab3.signal.spikes
        channel : {str, int}, optional
            Channel label or index in imaging dataset
        label : str, optional
            ROI label in imaging dataset. If None, the most recently accessed
            label in the imaging dataset will be used
        from_type : str {'dfof', 'raw'} or custom, optional
            Signal type to perform inference on. Defaults to 'dfof'
        overwrite : bool, optional
            Whether to overwrite signal key if it exists already. Defaults to
            False

        Returns
        -------
        spikes : pd.DataFrame
        """

        # TODO should we check to make sure this is a SpikesStrategy?
        assert strategy.name == 'spikes'
        return self.calculate_signals(strategy, to_type='spikes', **kwargs)

    def calculate_dfof(self, strategy, **kwargs):
        """Calculate dF/F from raw signals.

        Parameters
        ----------
        strategy : DFOFStrategy
            Instance of a DFOFStrategy. See lab3.signal.dfof
        channel : {str, int}, optional
            Channel label or index in imaging dataset
        label : str, optional
            ROI label in imaging dataset. If None, the most recently accessed
            label in the imaging dataset will be used
        overwrite : bool, optional
            Whether to overwrite signal key if it exists already. Defaults to
            False

        Returns
        -------
        dfof : pd.DataFrame
        """

        # TODO should we check to make sure this is a DFOFStrategy?
        assert strategy.name == 'DFOF'
        return self.calculate_signals(strategy, to_type='dfof', **kwargs)

    @default_label_handling
    def roi_list(self, label=None):
        """Get a list of ROI objects

        Parameters
        ----------
        label : str, optional
            Label of the desired ROI list. If None, the most recently accessed
            label in the imaging dataset will be used

        Returns
        --------
        sima.ROI.ROIList
            ROIList object containing all the ROIs for the passed label
        """

        # TODO default handling? or does SIMA handle this?
        return self.imaging_dataset.ROIs[label]

    def delete_roi_list(self, channel='Ch2', label=None, do_nothing=True):
        """Remove an existing ROI list. This both removes the ROIs from the
        underlying sima ImagingDataset (i.e. deletes them from
        folder.sima/rois.pkl), and also deletes any corresponding entries in
        signals.h5.

        Parameters
        ----------
        channel : {str, int}, optional
        label : str, optional
        do_nothing : bool, optional
        """

        if label in self.imaging_dataset.ROIs.keys():
            signals_file = self.signals_file(mode='a')
            keys_to_delete = [k for k in signals_file.keys()
                              if f'/{label}/' in k]
            print(f"Deleting keys:\n{keys_to_delete} "
                  + f"from {signals_file._path}")

            if not do_nothing:
                for key in keys_to_delete:
                    signals_file.remove(key)
                self.imaging_dataset.delete_ROIs(label=label)

            signals_file.close()

        else:
            raise KeyError(f"Label '{label}' is not in the imaging dataset")

    @default_label_handling
    def calculate_transients(self, channel='Ch2', label=None, overwrite=False,
                             strategy=None):
        raise NotImplementedError


class ImagingExperiment(ImagingMixin, BehaviorExperiment):

    """Imaging experiment with behavior. If the imaging directory is not
    already paired with the database trial id, or you would like to modify it,
    this can be passed during initialization and the database will be updated
    accordingly.

    Parameters
    ----------
    trial_id : int
        ID of experiment in the sql database
    sima_path : str, optional
        Path to a sima folder. If passed, this imaging data will be paired in
        the database with this trial_id.
    force_pairing : bool, optional
        If trial_id is already paired with a sima folder but sima_path
        is passed, whether to overwrite pairing. Default is False
    store : bool, optional
        Whether to store the newly paired sima_path in the database, if passed.
        Defaults to False.
    """

    def __init__(self, trial_id, sima_path=None, force_pairing=False,
                 store=False):

        super().__init__(trial_id)

        if sima_path is not None:
            try:
                if self._trial_info['sima_path'] and not force_pairing:
                    raise ValueError(
                        f"Trial {trial_id} is already paired to imaging"
                        + f" data {self._trial_info['sima_path']}. "
                        + "Set force_pairing=True to force pairing")
            except (KeyError, AttributeError):
                pass
            self.sima_path = sima_path
            self.save(store=store)
        else:
            try:
                self.sima_path = self._trial_info['sima_path']
                self.imaging_dataset
            except Exception:
                raise AttributeError(
                    f"Trial {trial_id} has no sima_path, or the path "
                    + "is not valid. Try passing a sima_path when "
                    + f"initializing {type(self).__name__}.")

    def format_behavior_data(self, image_sync=True, **kwargs):
        """Format behavior dictionary. Interval variables are converted to
        boolean vectors that denote activity at each sampling period.
        Continuous variables are re-sampled at each discrete sampling period.
        By default, behavior is binned to synchronize with the imaging data.

        Parameters
        ----------
        image_sync : bool, optional
            If True, sampling_interval is set to imaging frame period, and
            the length of behavior and imaging data are matched. Defaults to
            True.
        sampling_interval : float, optional
            Sampling rate for discretizing time. Variables expressed as
            intervals (i.e. vectors of start and stop times) are converted to
            binary vectors that are True for frames inside the intervals.
            Continuous variables are resampled at the corresponding time
            points. Ignored if image_sync is True.
        discard_pre : {'first_lap', float}, optional
            If 'first_lap', truncate data before the first complete lap. If
            float, truncate data occuring before this time. Default behavior
            returns all data points. Ignored if image_sync is True.
        discard_post : float, optional
            Truncate data occuring after this time. Default behavior returns
            all data points. Ignored if image_sync is True.
        sigma : float, optional
            Standard deviation of gaussian smoothing kernel for smoothing
            velocity variable, in seconds. Defaults to 0.1

        Returns
        -------
        beh_dict : dict
            Behavior data dictionary with variables expressed in discrete time
            bins

        See also
        --------
        experiment.utils.discretize_behavior
        """

        if image_sync:
            sigma = kwargs.pop('sigma', 0.1)
            if len(kwargs):
                warnings.warn("image_sync is True (default), so additional "
                              + "keyword arguments were ignored.")

            kwargs.update({'sampling_interval': self.frame_period,
                           'discard_post': self.frame_period
                           * self.imaging_dataset.num_frames,
                           'discard_pre': 0,
                           'sigma': sigma})

        return {**super().format_behavior_data(**kwargs),
                **{'image_sync': image_sync}}

    def velocity(self, image_sync=True, sigma=0.1, **beh_kws):
        """Get velocity trace.

        Parameters
        ----------
        image_sync : bool, optional
            If True, sampling_interval is set to imaging frame period, and
            the length of behavior and imaging data are matched. Defaults to
            True.
        sigma : float, optional
            Standard deviation of the gaussian smoothing kernel to apply to the
            velocity trace, in seconds. Defaults to 0.1.

        Additional keyword arguments are passed directly to
        format_behavior_data, and are used to override defaults only if
        image_sync is False.

        Returns
        -------
        velocity : array (n_samples,)
        """

        return super().velocity(image_sync=image_sync, sigma=sigma, **beh_kws)

    def signals(self, max_frame='trim_to_behavior', **kwargs):
        """Retrieve signals from signals.h5. By default, only return frames
        recorded with behavior data.

        Parameters
        ----------
        max_frame : int or 'trim_to_behavior', optional
            Only include the first max_frame frames. If 'trim_to_behavior',
            trim the excess frames to match behavior data length. Defaults to
            'trim_to_behavior'.
        channel : {str, int}, optional
            Channel name or index in the imaging dataset
        label : str, optional
            ROI list label in the imaging dataset
        signal_type : str {'raw', 'npil', 'dfof', 'spikes', or custom}
            Name of signal type, defaults to 'raw'

        Returns
        -------
        signals : pd.DataFrame
        """

        if max_frame == 'trim_to_behavior':
            # figure out length of behavior data in imaging frames
            beh_dict = self.format_behavior_data()
            max_frame = int(beh_dict['recordingDuration'] / self.frame_period)

        return super().signals(max_frame=max_frame, **kwargs)


class ImagingOnlyExperiment(ImagingMixin, _BaseExperiment):

    """Imaging experiment with no behavior / database records. To initialize,
    pass a SIMA directory rather than a trial ID.

    Parameters
    ----------
    sima_path : str
    """

    def __init__(self, sima_path):

        super().__init__(trial_id=sima_path)
        self.sima_path = sima_path


class LFPMixin:

    """Mixin for LFP experiments"""

    @property
    def lfp_path(self):
        """Return the location of the LFP data"""
        return self._lfp_path

    @lfp_path.setter
    def lfp_path(self, path):
        self._lfp_path = path

    @property
    def lfp_signals_path(self):
        return os.path.join(self.lfp_path, "lfp_signals.h5")

    def lfp_signals_file(self, mode='r', **kwargs):
        return SignalFile(self.lfp_signals_path, mode=mode, **kwargs)

    def lfp_signals(self, label, channel=None):
        """Retrieve signals from lfp_signals.h5.

        Parameters
        ----------
        label : str, optional
            Something like 'LFP', 'EMG', etc.
        channel : {str, int}, optional
            Channel name or index in the LFP dataset
        signal_type : str {'raw', 'filtered', or custom}
            Name of signal type, defaults to 'raw'

        Returns
        -------
        signals : pd.DataFrame
        """

        with self.lfp_signals_file() as signal_file:
            signals = signal_file.get(f"{label}")
        if channel is not None:
            signals = signals.iloc[channel]
        return signals

    @property
    def lfp_events_path(self):
        return os.path.join(self.lfp_path, 'lfp_events.h5')

    def lfp_events(self, event_type, label='LFP', channel=None, **kwargs):
        return LFPEventGroup.load(self, event_type=event_type,
                                  label=label, channel=channel, **kwargs)

    def check_overwrite(self, store_key, overwrite):
        with self.lfp_signals_file() as signals_file:
            key_exists = store_key in signals_file
            assert not key_exists or overwrite, \
                f"Key {store_key} already exists in signal file. Set " \
                + "overwrite=True to recalculate"


class LFPExperiment(LFPMixin, BehaviorExperiment):

    """LFP experiment with behavior but no imaging data"""

    def __init__(self, trial_id):
        super().__init__(trial_id)


class LFPOnlyExperiment(LFPMixin, _BaseExperiment):

    """LFP expreriment with no behavior or imaging data"""

    def __init__(self, lfp_path):
        super().__init__(trial_id=lfp_path)
        self.lfp_path = lfp_path


class ImagingLFPExperiment(LFPMixin, ImagingExperiment):

    """Combination 2p imaging and LFP experiment with behavior"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.lfp_path is None:
            self.lfp_path = os.path.join(os.path.split(
                                self.sima_path)[0], 'LFP')

    @property
    def imaging_times(self, **kwargs):
        """Attempt to timelock to LFP

        """
        nframes = self.signals().shape[1]

        try:
            frames = self.events(signal_type='LFP', event_type='imaging_frame')
            return np.array(frames.onset)[:nframes]
        except Exception:  # what error are we catching here?!
            print("Failed to timelock to LFP!")
            return super().imaging_times(**kwargs)


class ImagingLFPOnlyExperiment(LFPMixin, ImagingOnlyExperiment):
    pass


# for populating the documentation website
__all__ = [
    'BehaviorExperiment',
    'ImagingMixin',
    'ImagingExperiment',
    'ImagingOnlyExperiment',
    'LFPMixin',
    'LFPExperiment',
    'LFPOnlyExperiment',
    'ImagingLFPExperiment',
    'ImagingLFPOnlyExperiment',
    'fetch_trials',
]



from abc import ABCMeta, abstractmethod

import numpy as np
import pandas as pd
from scipy.signal import gaussian, convolve

from ..base import BaseSignalAnalysis


class SpatialTuningStrategy(BaseSignalAnalysis, metaclass=ABCMeta):

    """Abstract base class for implementing strategies for calculating
    spatial tuning. Each derived class should implement a specific strategy for
    calculating spatial tuning curves for each neuron / lap, and specify
    a method for shuffling position and/or activity to generating null
    distributions.

    The details here are flexible: in the most straightforward scenarios,
    the input signals is a DataFrame with shape (n_rois, n_samples), but it
    could also be a structure containing discrete events (e.g. transients). The
    output of _calculate however should always result in a DataFrame as
    described below.

    Implement _smooth, _shuffle, and _calculate to subclass. Optionally
    implement calculate_rasters to return trial-by-trial tuning curves.
    """

    def calculate(self, signals, position, laps=None, max_position=None,
                  shuffle=False):
        """For each neuron, calculate the spatial tuning curve averaged over
        laps. This method returns a dataframe that is indexable like:

        >>> tuning_curves.loc[roi_label, position_number]

        Parameters
        ----------
        signals : pd.DataFrame
            Typically an array shape (n_rois, n_samples), keyed by the ROI
            label.
        position : array (n_samples)
            Array containing the integer position bin for every sample in
            signals. The first bin on the track should be zero.
        laps : array (n_samples)
            Array containing the lap id of ever sample in signals.
        max_position : int, optional
            The maximum position bin on the track. If None (default), this is
            assumed to be max(position). In most cases where positions are
            densely sampled over the recording, this argument can be ignored.
            After calculating tuning curves and before smoothing, it is used
            to populate any unsampled position bins in the tuning curve with
            NaNs.
        shuffle : bool, optional
            Whether to shuffle the position of each sample before computing
            tuning curves (i.e. for sampling the tuning curve null
            distribution)

        Returns
        -------
        tuning_curves : pd.DataFrame (n_rois, n_position_bins)
            Dataframe storing the spatial tuning curve for each cell.
        """

        if shuffle:
            signals, position, laps = self._shuffle(
                signals=signals, position=position, laps=laps)
        tuning_curves = self._calculate(signals, position, laps)

        # insert NaNs for missing position bins in the tuning curve
        if max_position is None:
            max_position = np.max(position)

        missing_bins = []
        for n in range(max_position):
            if n not in tuning_curves.columns:
                missing_bins.append(n)
                tuning_curves.loc[:, n] = np.nan
        tuning_curves = tuning_curves.sort_index(axis=1)

        if missing_bins:
            # interpolate to replace missing values for now before smoothing
            smooth_tuning_curves = self._smooth(
                tuning_curves.interpolate(
                    axis=1, method='linear', limit_direction='both'))
            # re-insert nans at missing values
            for n in missing_bins:
                smooth_tuning_curves.loc[:, n] = np.nan
            return smooth_tuning_curves
        else:
            return self._smooth(tuning_curves)

    # ----------------- Implement these methods to subclass ----------------- #

    @abstractmethod
    def _smooth(self, tuning_curves):
        """Apply smoothing to the resulting tuning curves

        Parameters
        ----------
        tuning_curves : pd.DataFrame (n_rois, n_position_bins)
            Dataframe storing the spatial tuning curve for each cell.

        Returns
        -------
        tuning_curves : pd.DataFrame (n_rois, n_position_bins)
            Dataframe storing the spatial tuning curve for each cell.
        """

    @abstractmethod
    def _shuffle(self, signals=None, position=None, laps=None):
        """Shuffle the data. For generating null tuning curves, you will
        typically want to shuffle either the signals themselves or their
        associated positions.

        Parameters
        ----------
        signals : array (n_rois, n_samples), optional
            Typically an array shape (n_rois, n_samples), keyed by the ROI
            label.
        position : array (n_samples), optional
            Array containing the integer position bin for every sample in
            signals
        laps : array (n_samples), optional
            Array containing the lap id of ever sample in signals. Depending on
            your shuffling method, you may or many not need the laps variable.
        """

    @abstractmethod
    def _calculate(self, signals, position, laps=None):
        """Calculate spatial tuning for each neuron. This should return a
        dataframe indexable like:

        >>> tuning_curves.loc[roi_label, position_number]

        Parameters
        ----------
        signals : array (n_rois, n_samples)
            Typically an array shape (n_rois, n_samples), keyed by the ROI
            label.
        position : array (n_samples)
            Array containing the integer position bin for every sample in
            signals
        laps : array (n_samples), optional
            Array containing the lap id of ever sample in signals. Depending on
            your tuning calculation method, you may or many not need the laps
            variable.
        """

    # ------------------------------ Optional ------------------------------- #

    def calculate_rasters(self, signals, position, laps, smooth=True):
        """This should return a multi-index dataframe with the spatial tuning
        for each lap, indexible like:

        >>> rasters.loc[(roi_label, lap_number), position_number]

        Depending on the algorithm, it may or may not be performant to use
        calculate_rasters when implementing _calculate. It is still useful to
        implement anyway, for example, to provide trial-level tuning for
        plotting.

        Parameters
        ----------
        signals : array (n_rois, n_samples)
            Typically an array shape (n_rois, n_samples), keyed by the ROI
            label.
        position : array (n_samples)
            Array containing the integer position bin for every sample in
            signals
        laps : array (n_samples)
            Array containing the lap id of ever sample in signals. Depending on
            your tuning calculation method, you may or many not need the laps
            variable.
        smooth : bool, optional
            Whether to apply smoothing to each lap. Defaults to True
        """
        raise NotImplementedError


class SimpleSpatialTuning(SpatialTuningStrategy):

    """A simple spatial tuning strategy. For each neuron, we compute its
    average activity across samples/laps for each position. The tuning curves
    are optionally smoothed with a Gaussian kernel. Since we are computing
    the average activity across samples for a given position, this implicitly
    normalizes for occupancy.

    Shuffled curves are generated by circularly shifting the position of
    samples independently within each lap.

    Parameters
    ----------
    sigma : float, optional
        Standard deviation of the gaussian smoothing kernel applied to the
        tuning curves. By default, no smoothing is applied.

    Examples
    --------
    >>> # generate some fake data
    >>> signals = pd.DataFrame(np.random.normal(0, 1, (10, 1000)))
    >>> laps = np.concatenate([np.ones(100) * lap for lap in range(10)])
    >>> position = np.concatenate([np.arange(100)] * 10)
    >>>
    >>> # compute the true tuning curves
    >>> strategy = SimpleSpatialTuning(sigma=2)
    >>> tuning_curves = strategy.calculate(
    >>>     signals, position, laps, shuffle=False)
    >>>
    >>> # get a set of null tuning curves via circular shuffle
    >>> tuning_curves = strategy.calculate(
    >>>     signals, position, laps, shuffle=True)
    """

    def __init__(self, sigma=None):
        self.sigma = sigma
        if self.sigma:
            kernel = gaussian(sigma * 8, sigma)
            self._k = kernel / kernel.sum()

    def _smooth(self, tuning_curves, random_state=None):
        """Apply Gaussian smoothing across position bins"""
        if self.sigma is None:
            return tuning_curves
        else:
            tuning_curves.iloc[:, :] = np.apply_along_axis(
                convolve, 1, tuning_curves, self._k, mode='same')
            return tuning_curves

    def _shuffle(self, signals=None, position=None, laps=None):
        """Applies a circular shuffle. On ever lap, shuffle the position
        samples by a random number of samples. signals input is ignored."""
        shuffled_position = np.concatenate([np.roll(
            position[(laps == l)], np.random.randint((laps == l).sum()))
                               for l in np.unique(laps)])
        return signals, shuffled_position, laps

    def _calculate(self, signals, position, laps=None):
        """For ROI, simply calculate the average of all samples within each
        position.

        Parameters
        ----------
        signals : pd.DataFrame
        position : array
        laps : ignored
        """

        columns = pd.MultiIndex.from_arrays([position, signals.columns],
                                            names=['position', 'time'])
        signals = pd.DataFrame(signals.values, columns=columns,
                               index=pd.Index(signals.index, name='roi_label'))
        return signals.T.groupby('position').mean().T

    def calculate_rasters(self, signals, position, laps, smooth=True,
                          fill_nans=True):
        """doc string

        Parameters
        ----------
        signals : pd.DataFrame
        position : array
        laps : array
        fill_nans : bool
        """

        columns = pd.MultiIndex.from_arrays([position, laps],
                                            names=['position', 'lap'])
        signals = pd.DataFrame(signals.values, columns=columns,
                               index=pd.Index(signals.index, name='roi_label'))
        rasters = signals.T.groupby(
            ['position', 'lap']).mean().unstack('lap').T

        if fill_nans:
            rasters = rasters.fillna(0)
        if smooth:
            rasters = self._smooth(rasters)
        return rasters


class CircularMixin:

    """Mixin class that modifies the functionality of _smooth, in order to
    account for the circularity of the track. In derived classes, this should
    be the first parent class.
    """

    def _smooth(self, tuning_curves):
        n_position_bins = tuning_curves.shape[1]
        tuning_curves = pd.concat([tuning_curves] * 3, axis=1)
        smooth_curves = super()._smooth(tuning_curves)
        return smooth_curves.iloc[:, n_position_bins:(n_position_bins * 2)]

    def _trial_smooth(self, rasters):
        n_position_bins = rasters.shape[1]

        # iterate over ROIs and append the previous and next lap to the left
        # and right of each trace respectively
        trial_curves = []
        for label, roi in rasters.groupby('roi_label'):
            previous_lap = np.vstack([np.zeros(roi.shape[1]),
                                      roi.values[:-1, :]])
            next_lap = np.vstack([roi.values[1:, :],
                                  np.zeros(roi.shape[1])])
            trial_curves.append(pd.DataFrame(
                np.hstack([previous_lap, roi.values, next_lap]),
                index=roi.index))

        # smooth
        smooth_curves = super()._smooth(pd.concat(trial_curves))
        smooth_curves = \
            smooth_curves.iloc[:, n_position_bins:(n_position_bins * 2)]
        smooth_curves.columns = rasters.columns
        return smooth_curves

    def calculate_rasters(self, *args, smooth=True, **kwargs):
        rasters = super().calculate_rasters(*args, smooth=False, **kwargs)
        if smooth:
            return self._trial_smooth(rasters)
        else:
            return rasters


class CircularSimpleSpatialTuning(CircularMixin, SimpleSpatialTuning):

    """Tuning strategy identical to SimpleSpatialTuning, but with smoothing
    that accounts for the edge effects of circular tracks.

    Parameters
    ----------
    sigma : float, optional
        Standard deviation of the gaussian smoothing kernel applied to the
        tuning curves. By default, no smoothing is applied.
    """

    pass


class TransientSpatialTuning(SpatialTuningStrategy):

    """TODO : the format of this will depend on the manner in which transients
    are stored.
    """

    pass


class CircularTransientSpatialTuning(CircularMixin, TransientSpatialTuning):

    pass



import numpy as np 
import scipy as sp
import scipy.ndimage

import scipy.stats as stats

def nan_zscore(array):
    """Compute the zscore of an array which may contain nans.
    Parameters
    ----------
    array : array-like
        Array to compute zscores on
    """
    return (array - np.nanmean(array))/np.nanstd(array)

def nonparametric_nan_zscore(array):
    """Compute the nonparametric zscore of an array which may contain nans.
    Parameters
    ----------
    array : array-like
        Array to compute nonparametric zscores on. Use median and IQR 
        instead of mean and std. Default to False.
    """
    return (array - np.nanmedian(array))/stats.iqr(array, nan_policy='omit')    

def nan_filter(array, filter_func, **filter_kwargs):
    """Implements a normalized convolution. This is only well-defined
    for LINEAR (i.e., convolution-based) filters filter_func.
    
    Example
    ----------
    >>> U=sp.randn(10,10)        
    >>> U[U>2]=np.nan              
    >>> U_filtered = nan_filter(U, sp.ndimage.gaussian_filter, 
                            sigma=2, truncate=4)
    """

    V = np.nan_to_num(array.copy())
    V_filtered = filter_func(V, **filter_kwargs)

    W = np.ones_like(array)
    W[np.isnan(array)] = 0
    W_filtered = filter_func(W, **filter_kwargs)

    return V_filtered/W_filtered



